# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from autologging import logged

from kqcircuits.elements.element import Element
from kqcircuits.util.geometry_helper import circle_polygon
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.defaults import default_bump_parameters


@logged
class FlipChipConnector(Element):
    """Connector between matching faces of two chips.

    The connector makes a galvanic contact between the flipped over top chip and the bottom chip.
    Origin is at the geometric center.
    """

    ubm_diameter = Param(pdt.TypeDouble, "Under-bump metalization diameter",
                         default_bump_parameters['under_bump_diameter'], unit="μm")
    bump_diameter = Param(pdt.TypeDouble, "Bump diameter", default_bump_parameters['bump_diameter'], unit="μm")

    def create_bump_connector(self):
        ubm_shape = circle_polygon(self.ubm_diameter / 2, self.n)
        self.cell.shapes(self.get_layer("underbump_metallization", 0)).insert(ubm_shape)
        self.cell.shapes(self.get_layer("underbump_metallization", 1)).insert(ubm_shape)

        avoidance_shape = circle_polygon(self.ubm_diameter / 2 + self.margin, self.n)
        self.cell.shapes(self.get_layer("ground_grid_avoidance", 0)).insert(avoidance_shape)
        self.cell.shapes(self.get_layer("ground_grid_avoidance", 1)).insert(avoidance_shape)

        bump_shape = circle_polygon(self.bump_diameter / 2, self.n)
        self.cell.shapes(self.get_layer("indium_bump", 0)).insert(bump_shape)  # bottom In bump
        self.cell.shapes(self.get_layer("indium_bump", 1)).insert(bump_shape)  # top In bump



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


"""Configuration file for KQCircuits.

Defines values for things such as default layers, paths, and default sub-element types.
"""
import os
import subprocess
from pathlib import Path

from kqcircuits.pya_resolver import pya
from kqcircuits.util.import_helper import module_from_file


_kqcircuits_path = Path(os.path.dirname(os.path.realpath(__file__)))
# workaround for Windows because os.path.realpath doesn't work there before Python 3.8
if os.name == "nt" and os.path.islink(Path(__file__).parent):
    _kqcircuits_path = Path(os.readlink(str(Path(__file__).parent)))

# project paths
SRC_PATHS = [_kqcircuits_path]
ROOT_PATH = Path(os.getenv('KQC_ROOT_PATH', os.getcwd()))  # "current dir" or set by optional KQC_ROOT_PATH
if _kqcircuits_path.parts[-3] == "klayout_package":  # developer setup
    ROOT_PATH = _kqcircuits_path.parents[2]

TMP_PATH = Path(os.getenv('KQC_TMP_PATH', ROOT_PATH.joinpath("tmp")))  # specify alternative tmp directory
_py_path = ROOT_PATH.joinpath("klayout_package/python")

if _kqcircuits_path.parts[-4] == "salt":  # KQC Salt package
    ROOT_PATH = _kqcircuits_path.parents[1]
    _py_path = ROOT_PATH.joinpath("python")
    TMP_PATH = _kqcircuits_path.parents[3].joinpath("python/tmp")  # local tmp dir for salt package

SCRIPTS_PATH = _py_path.joinpath("scripts")
DRC_PATH = _py_path.joinpath("drc")

TMP_PATH.mkdir(parents=True, exist_ok=True)  # TODO move elsewhere?

ANSYS_SCRIPT_PATHS = [SCRIPTS_PATH.joinpath("simulations").joinpath("ansys")]
ELMER_SCRIPT_PATHS = [SCRIPTS_PATH.joinpath("simulations").joinpath("elmer")]
XSECTION_PROCESS_PATH = ROOT_PATH.joinpath("xsection/kqc_process.xs")

VERSION_PATHS = {}
VERSION_PATHS['KQC'] = ROOT_PATH
SIM_SCRIPT_PATH = ROOT_PATH / 'klayout_package' / 'python' / 'scripts' / 'simulations'

# Given to subprocess.Popen calls, hides terminals on Windows
STARTUPINFO = None
if os.name == "nt":
    STARTUPINFO = subprocess.STARTUPINFO()
    STARTUPINFO.dwFlags |= subprocess.STARTF_USESHOWWINDOW

# printed to corners of all chips and top of all masks
# could be for example "IQM" or "A!"
default_brand = "IQM"

# default output format and extension
default_output_format = "OASIS"

# default bitmap dimensions
default_png_dimensions = (1000, 1000)

default_junction_type = "Manhattan"
default_airbridge_type = "Airbridge Rectangular"
default_fluxline_type = "Fluxline Standard"
default_marker_type = "Marker Standard"
default_junction_test_pads_type = "Junction Test Pads Simple"
default_tsv_type = "Tsv Standard"

default_drc_runset = "example.lydrc"

# default elements to breakdown before netlist export
# list of strings
default_netlist_breakdown = [
    "Waveguide Composite",
    "Meander",
]

default_netlist_ignore_connections = [
    ("drive", "drive"), # Don't connect two overlapping qubit drive ports
]

# default progress bar formatting with tqdm
default_bar_format = '{l_bar}{bar}| {n_fmt}/{total_fmt} [Elapsed: {elapsed}, Left (eta): {remaining}, {rate_inv_fmt}' \
                     '{postfix}]'

# dictionary of probepoint types for probepoint export
# key is the refpoint name prefix for a type of probepoints, value is a text representation of the probepoint type
default_probe_types = {"testarray": "testarrays", "qb": "qubits"}
# tuple of refpoint name suffixes that are used (together with probe_types) to identify refpoints as probepoints
default_probe_suffixes = ("_c", "_l")

# Library names in dependency order. Every library should have its dependencies before its own entry.
kqc_library_names = (
    'Element Library',
    'Junction Library',
    'Test Structure Library',
    'Qubit Library',
    'Chip Library',
)

# The user may override KQC Element's default parameters
default_parameter_values = {
    "AirbridgeConnection": {"bridge_length": 60, "pad_length": 22},
    "TsvTest": {"tsv_elliptical_width": 10},
}

default_bump_parameters = {
    "bump_diameter": 25,
    "under_bump_diameter": 40,
    "bump_grid_spacing": 120,
    "bump_edge_to_bump_edge_separation": 95,
    "edge_from_bump": 550,
}


# Dictionary of sample holders to determine launcher parameters and chip sizes. Keys are sampleholder names, values are
# dictionaries containing the following items:
#
#            n: number of launcher pads or an array of pad numbers per side
#            launcher_type: type of the launchers, "RF" or "DC"
#            launcher_width: width of the launchers
#            launcher_gap: pad to ground gap of the launchers
#            launcher_indent: distance between the chip edge and pad port
#            pad_pitch: distance between pad centers
#            chip_box: chip size
#
default_sampleholders = {
    "SMA8": {
        "n": 8,
        "launcher_type": "RF",
        "launcher_width": 300,
        "launcher_gap": 260,
        "launcher_indent": 800,
        "launcher_frame_gap": 180,
        "pad_pitch": 4400,
        "chip_box": pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000))
    },
    "ARD24": {
        "n": 24,
        "launcher_type": "RF",
        "launcher_width": 240,
        "launcher_gap": 179,
        "launcher_indent": 680,
        "pad_pitch": 1200,
        "chip_box": pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000))
    },
    "DC24": {
        "n": 24,
        "launcher_type": "DC",
        "launcher_width": 500,
        "launcher_gap": 300,
        "launcher_indent": 680,
        "pad_pitch": 850,
        "chip_box": pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000))
    }
}

# Elements that are visible in the Node editor plugin dropdown box, specified by class name. A list of in principle
# valid elements can be generated with kqcircuits.util.gui_helper.get_valid_node_elements, but that function generates
# an instance of each PCell which is clumsy to run at startup.
# In the list below basic waveguide elements are omitted since they can be better generated by other node parameters
node_editor_valid_elements = [
    'Airbridge', 'AirbridgeConnection', 'AirbridgeMultiFace', 'AirbridgeRectangular', 'CircularCapacitor',
    'FingerCapacitorSquare', 'FingerCapacitorTaper', 'FlipChipConnectorRf', 'SmoothCapacitor',
    'WaveguideCoplanarSplitter']

# Path to the layer configuration file, which defines layer/face related defaults.
# The path can be either absolute or relative.
layer_config_path = Path(__file__).parent/"layer_config"/"default_layer_config.py"

# Load layer/face related defaults from the layer config file

layer_config_module = module_from_file(layer_config_path)
# pya layer information
default_layers = layer_config_module.default_layers
# default_faces[face_id] contains the face dictionary for the face determined by face_id.
#
# Each face dictionary should contain:
#   - key "id" with value face_id (string)
#   - for all the available layers in that face: key "Layer_name", value pya.LayerInfo object for that layer
default_faces = layer_config_module.default_faces
# face_id of the face that is used by default in some contexts
default_face_id = layer_config_module.default_face_id
# Layer names (without face prefix) for layers exported as individual .oas files during mask layout export
default_mask_export_layers = layer_config_module.default_mask_export_layers
# Layer names (without face prefix) with mask label postfix for mask label and mask covered region creation
default_layers_to_mask = layer_config_module.default_layers_to_mask
# Layer names (without face prefix) in `layers_to_mask` for which mask covered region is not created
default_covered_region_excluded_layers = layer_config_module.default_covered_region_excluded_layers
# Layer names (without face prefix) for layers exported as bitmap files during full mask layout export (does not
# apply to individual pixels)
mask_bitmap_export_layers = layer_config_module.mask_bitmap_export_layers
# Layers to hide when exporting a bitmap with "all" layers
all_layers_bitmap_hide_layers = layer_config_module.all_layers_bitmap_hide_layers
# Layer clusters used for exporting only certain layers together in the same file, when exporting individual chips
# during mask layout export (dict with items `cluster name: LayerCluster`)
chip_export_layer_clusters = layer_config_module.chip_export_layer_clusters
# Default layers to use for calculating cell path lengths with get_cell_path_length()
default_path_length_layers = layer_config_module.default_path_length_layers
# Default mask parameters for each face (dict with items `face_id: parameters`)
default_mask_parameters = layer_config_module.default_mask_parameters
# Path to layer properties file
default_layer_props = layer_config_module.default_layer_props
# Prefixes to use in chip name labels for different faces (dict with items `face_id: label_prefix`)
default_chip_label_face_prefixes = layer_config_module.default_chip_label_face_prefixes
# Update default_parameter_values based on layer config file
for k, v in layer_config_module.default_parameter_values.items():
    params = default_parameter_values[k] if k in default_parameter_values else {}
    for k2, v2 in v.items():
        params[k2] = v2
    default_parameter_values[k] = params



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.pya_resolver import pya
from kqcircuits.test_structures.test_structure import TestStructure
from kqcircuits.util.parameters import Param, pdt


class AirbridgeDC(TestStructure):
    """The PCell declaration for an airbridge test structure for four-point DC measurements."""

    n_ab = Param(pdt.TypeInt, "Number of airbridges", 30)
    pad_height = Param(pdt.TypeDouble, "Pad height", 500, unit="μm")
    width = Param(pdt.TypeDouble, "Total width", 2000, unit="μm")

    def build(self):

        # create airbridges and islands through which they are connected

        cell_ab = self.add_element(Airbridge, airbridge_type="Airbridge Rectangular")

        ab_params = cell_ab.pcell_parameters_by_name()
        bridge_width = ab_params["bridge_width"]
        ab_pad_extra = ab_params["pad_extra"]
        ab_pad_width = ab_params["pad_length"] - 2 * ab_pad_extra
        ab_pad_length = ab_params["pad_length"] - 2 * ab_pad_extra
        bridge_length = ab_params["bridge_length"] + 2 * ab_pad_extra

        island_margin = 5  # how much an island extends beyond airbridge pads
        island_width = bridge_width + ab_pad_extra*2 + 2*island_margin
        airbridge_separation = 30
        island_height = max(2*(ab_pad_length + ab_pad_width/2), 2*(ab_pad_width+ab_pad_extra)) + airbridge_separation\
                        + island_margin*2
        island = pya.DPolygon([
            pya.DPoint(0, 0),
            pya.DPoint(0, island_height),
            pya.DPoint(island_width, island_height),
            pya.DPoint(island_width, 0),
        ])

        islands_region = pya.Region()

        x_step = island_width + bridge_length
        y_step = island_height + bridge_length

        pad_spacing_y = 150

        # The airbridges are created from left to right, by "snaking" in such a way that minimal horizontal space is
        # used, while keeping within the vertical extents of the pads.

        n_ab_placed = 0
        n_ab_remaining = self.n_ab
        n_ab_horizontal = 0
        # The ab_type here depends the position and direction of the airbridge:
        # "bottom" and "top" for horizontal airbridges at top or bottom of a vertical sequence of airbridges
        # "up" and "down" for vertical airbridges depending on which direction the airbridge is compared to previous one
        ab_type = "bottom"
        x = 0
        y = 0
        row = 0

        while n_ab_remaining > 0:

            if ab_type == "bottom":
                ab_trans = pya.DTrans(1, False, x + (island_width + x_step)/2,
                                      y + island_margin + ab_pad_width/2 + ab_pad_extra)
                x += x_step
                if row == 0 and n_ab_remaining < 5:
                    ab_type = "top"
                else:
                    ab_type = "up"
                n_ab_horizontal += 1
            elif ab_type == "up":
                ab_trans = pya.DTrans(0, False, x + island_width/2,
                                      y + island_height + (y_step - island_height)/2)
                y += y_step
                row += 1
                if y + 2*island_height + bridge_length > pad_spacing_y/2 + self.pad_height + island_height/2 or \
                   (row >= 0 and (n_ab_remaining <= row + 4 or n_ab_remaining == row + 6)):
                    ab_type = "top"
            elif ab_type == "top":
                ab_trans = pya.DTrans(1, False, x + (island_width + x_step)/2,
                                      y + island_height - island_margin - ab_pad_width/2 - ab_pad_extra)
                x += x_step
                if row == 0 and n_ab_remaining < 5:
                    ab_type = "bottom"
                else:
                    ab_type = "down"
                n_ab_horizontal += 1
            else:  # ab_type == "down"
                ab_trans = pya.DTrans(0, False, x + island_width/2, y - (y_step - island_height)/2)
                y -= y_step
                row -= 1
                if y - island_height - bridge_length < -pad_spacing_y/2 - self.pad_height + island_height/2 or \
                   (row < 0 and (n_ab_remaining <= -row + 4 or n_ab_remaining == -row + 6)) or \
                   (row == 0 and n_ab_remaining < 5):
                    ab_type = "bottom"

            self.insert_cell(cell_ab, ab_trans)

            if n_ab_remaining > 1:
                island_trans = pya.DTrans(pya.DVector(x, y))
                islands_region.insert((island_trans*island).to_itype(self.layout.dbu))

            n_ab_placed += 1
            n_ab_remaining -= 1

        pad_spacing_x = n_ab_horizontal*(bridge_length + island_width) - island_width

        # once all the airbridges and islands have been created, transform them so that they are centered around (0, 0)
        islands_trans = pya.DTrans(pya.DVector(-pad_spacing_x/2 - island_width, -island_height/2))
        islands_region.transform(islands_trans.to_itype(self.layout.dbu))
        for inst in self.cell.each_inst():
            inst.transform(islands_trans)

        # create pads
        pads_region = pya.Region()
        gap_extra = 50  # how much the gap layer extends beyond the pads
        pad_width = (self.width - pad_spacing_x)/2 - gap_extra
        self.produce_four_point_pads(pads_region, pad_width, self.pad_height, pad_spacing_x, pad_spacing_y, True,
                                     refpoint_distance=100)

        # combine pads and islands and create the metal gap region based on them
        metal_region = islands_region + pads_region
        self.produce_etched_region(metal_region, pya.DPoint(0, 0), 2*pad_width + pad_spacing_x + 2*gap_extra,
                                   2*self.pad_height + pad_spacing_y + 2*gap_extra)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from kqcircuits.simulations.simulation import Simulation
from kqcircuits.pya_resolver import pya
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare
from kqcircuits.util.parameters import add_parameters_from


@add_parameters_from(FingerCapacitorSquare)
class FingerCapacitorSim(Simulation):

    def build(self):
        capacitor_cell = self.add_element(FingerCapacitorSquare)

        cap_trans = pya.DTrans(0, False, (self.box.left + self.box.right) / 2, (self.box.bottom + self.box.top) / 2)
        _, refp = self.insert_cell(capacitor_cell, cap_trans)

        a2 = self.a if self.a2 < 0 else self.a2
        b2 = self.b if self.b2 < 0 else self.b2
        self.produce_waveguide_to_port(refp["port_a"], refp["port_a_corner"], 1, 'left', a=self.a, b=self.b)
        self.produce_waveguide_to_port(refp["port_b"], refp["port_b_corner"], 2, 'right', a=a2, b=b2)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from autologging import logged

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.junctions.junction import Junction
from kqcircuits.util.symmetric_polygons import polygon_with_vsym


@logged
class Sim(Junction):
    """The PCell declaration for a simulation type SQUID.

    Origin at the center of junction layer bottom edge.
    """

    junction_total_length = Param(pdt.TypeDouble, "Simulation junction total length", 33, unit='µm')

    def build(self):

        trans = pya.DTrans(0.0, 0.0)
        self._produce_ground_metal_shapes(trans)
        # refpoints
        self.refpoints["origin_squid"] = pya.DPoint(0, 0)
        self.refpoints["port_squid_a"] = pya.DPoint(0, self.junction_total_length - 13)
        self.refpoints["port_squid_b"] = pya.DPoint(0, 12)
        self.refpoints["port_common"] = pya.DPoint(0, self.junction_total_length)


    def _produce_ground_metal_shapes(self, trans):
        """Produces hardcoded shapes in metal gap and metal addition layers."""
        # metal additions bottom
        bottom_pts = [
            pya.DPoint(-4, 0),
            pya.DPoint(-4, 12),
            pya.DPoint(4, 12),
            pya.DPoint(4, 0)
        ]
        shape = polygon_with_vsym(bottom_pts)
        self.cell.shapes(self.get_layer("base_metal_addition")).insert(trans*shape)
        # metal additions top
        top_pts = [
            pya.DPoint(-4, self.junction_total_length - 13),
            pya.DPoint(-4, self.junction_total_length),
            pya.DPoint(4, self.junction_total_length),
            pya.DPoint(4, self.junction_total_length - 13)
        ]
        shape = polygon_with_vsym(top_pts)
        self.cell.shapes(self.get_layer("base_metal_addition")).insert(trans*shape)
        # add ground grid avoidance
        w = self.cell.dbbox().width()
        h = self.cell.dbbox().height()
        protection = pya.DBox(-w / 2 - self.margin, - self.margin, w / 2 + self.margin, h + self.margin)
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(trans*protection)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from math import pi, cos, sin, tan, atan, atan2, degrees, sqrt
from scipy.optimize import brentq

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.element import Element
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.util.geometry_helper import vector_length_and_direction, get_angle


class Meander(Element):
    """The PCell declaration for a meandering waveguide.

    Defined by two points, total length and number of meanders.

    The start and end points can be moved in GUI using the "Move"-tool. Alternatively, if a list of ``[x, y]``
    coordinates is given for ``start`` and ``end``, the GUI markers will not be shown. The latter is useful for
    code-generated cells that cannot be edited in the GUI.

    By default, the number of meanders is automatically chosen to minimize the area taken by bounding box of
    the meander. Uses the same bending radius as the underlying waveguide. Equidistant airbridges can be placed in the
    meander using ``n_bridges`` parameter.

       .. MARKERS_FOR_PNG 0,0
    """
    start = Param(pdt.TypeShape, "Start", pya.DPoint(-600, 0))
    end = Param(pdt.TypeShape, "End", pya.DPoint(600, 0))
    length = Param(pdt.TypeDouble, "Length", 3000, unit="μm")
    meanders = Param(pdt.TypeInt, "Number of meanders (non-positive means automatic)", -1)
    n_bridges = Param(pdt.TypeInt, "Number of bridges", 0)

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def parameters_from_shape_impl(self):
        points = [pya.DPoint(point * self.layout.dbu) for point in self.shape.each_point()]
        self.start = points[0]
        self.end = points[-1]

    def build(self):
        if isinstance(self.start, list):
            start = pya.DPoint(self.start[0], self.start[1])
        else:
            start = self.start
        if isinstance(self.end, list):
            end = pya.DPoint(self.end[0], self.end[1])
        else:
            end = self.end

        angle = 180/pi*atan2(end.y - start.y, end.x - start.x)
        transf = pya.DCplxTrans(1, angle, False, pya.DVector(start))
        l_direct = start.distance(end)
        if l_direct < 4*self.r:
            self.raise_error_on_cell(
                "Cannot create a Meander because start and end points are too close to each other.",
                (start + end)/2)
        if self.meanders < 1:
            self.meanders = int(l_direct/(2*self.r) - 1)  # automatically choose maximum possible number of meanders

        target_increment = self.length - l_direct  # target length increment compared to straight segment
        if target_increment < -1e-3:
            self.raise_error_on_cell("Cannot create a Meander with the given parameters. Try increasing the length.",
                                     (start + end) / 2)

        def bend_corner_displacement(w):
            """Returns x-displacement of corner point as function of bend width w.

            Waveguide starts as straight segment from origin (0,0) and ends with bend pointing towards x at (self.r,w).
            """
            if w >= self.r:
                return 0.0
            return (self.r - w) / (1 - w / (2 * self.r))

        def bend_length_increment(w):
            """Returns amount of waveguide length increment due to a single bend as function of bend width w.

            Waveguide starts as straight segment from origin (0,0) and ends with bend pointing towards x at (self.r,w).
            """
            if w >= self.r:
                return self.r * (pi / 2 - 2) + w
            h = w / self.r
            x = (1 - h) / (1 - h / 2)
            return self.r * (2 * atan(1 - x) + (x + h * (h - 1)) / sqrt(x ** 2 + h ** 2) - 1)

        def meander_length_increment(w):
            """Returns amount of waveguide length increment due to all meander bends as function of meander width w."""
            l0 = bend_length_increment(w / 4)  # starting and ending bend increment
            l1 = bend_length_increment(w / 2)  # middle bend increment
            return 4 * l0 + 2 * (self.meanders - 1) * l1

        # Create meander points
        points = [pya.DPoint(0, 0)]
        if target_increment > 1e-3:
            min_90deg_increment = meander_length_increment(4 * self.r)  # minimal length increment for 90-degree bends
            if target_increment >= min_90deg_increment:  # if all meander bends are 90 degrees
                width = 4 * self.r + (target_increment - min_90deg_increment) / self.meanders
            else:  # computation of meander width is not trivial, so we need to use root finding algorithm
                width = brentq(lambda w: meander_length_increment(w) - target_increment, 0.0, 4 * self.r)

            l_rest = l_direct / 2 - self.r * self.meanders  # distance to first corner
            x0 = bend_corner_displacement(width / 4)  # x-displacement of corner points in both ends
            x1 = bend_corner_displacement(width / 2)  # x-displacement of corner points in the middle of meander

            points.append(pya.DPoint(l_rest - x0, 0))
            points.append(pya.DPoint(l_rest + x0, width/2))
            for i in range(1, self.meanders):
                x = l_rest + 2 * self.r * i
                points.append(pya.DPoint(x - x1, (-1)**(i+1) * width / 2))
                points.append(pya.DPoint(x + x1, (-1)**i * width / 2))
            points.append(pya.DPoint(l_direct - l_rest - x0, (-1)**(self.meanders+1) * width / 2))
            points.append(pya.DPoint(l_direct - l_rest + x0, 0))
        points.append(pya.DPoint(l_direct, 0))

        # Create airbridges
        if self.n_bridges > 0:
            self._produce_bridges(points, transf)

        # Insert waveguide and ports
        waveguide = self.add_element(WaveguideCoplanar, path=points)
        wg_inst, _ = self.insert_cell(waveguide, transf)
        self.copy_port("a", wg_inst)
        self.copy_port("b", wg_inst)

    def _produce_bridges(self, points, trans):
        """Produces equally spaced airbridges on top of the meander waveguide.

        Args:
            points: list of points for the waveguide path
            trans: transformation applied to the entire meander
        """

        def insert_bridge(position, angle):
            self.insert_cell(Airbridge, trans*pya.DCplxTrans(1, angle, False, position))

        bridge_separation = self.length / (self.n_bridges + 1)
        dist_to_next = bridge_separation
        n_inserted = 0

        # Insert airbridges on bends and between bends
        for i in range(1, len(points)-1):
            v1, _, a1, a2, c_pos = WaveguideCoplanar.get_corner_data(points[i - 1], points[i], points[i + 1], self.r)
            alpha = (a2 - a1 + pi) % (2 * pi) - pi  # turn angle (between -pi and pi) in radians
            sign_r = (self.r if alpha > 0 else -self.r)  # positive or negative radius depending on turn signature
            length, direction = vector_length_and_direction(v1)
            cut_dist = self.r * tan(abs(alpha) / 2)  # distance between corner point and beginning of straights

            dist_to_next -= length - cut_dist
            while dist_to_next <= 0.0:  # insert airbridges on straight segment before corner
                insert_bridge(points[i] + (dist_to_next - cut_dist) * direction, degrees(a1))
                dist_to_next += bridge_separation
                n_inserted += 1
                if n_inserted >= self.n_bridges:
                    return
            dist_to_next -= alpha * sign_r
            while dist_to_next <= 0.0:  # insert airbridges on corner segment
                a = a2 + dist_to_next / sign_r
                insert_bridge(c_pos + sign_r * pya.DVector(sin(a), -cos(a)), degrees(a))
                dist_to_next += bridge_separation
                n_inserted += 1
                if n_inserted >= self.n_bridges:
                    return
            dist_to_next += cut_dist

        # Insert airbridges on the last segment
        length, direction = vector_length_and_direction(points[-1] - points[-2])
        dist_to_next -= length
        while dist_to_next <= 0.0:  # insert airbridges on last straight segment
            insert_bridge(points[-1] + dist_to_next * direction, get_angle(direction))
            dist_to_next += bridge_separation
            n_inserted += 1
            if n_inserted >= self.n_bridges:
                return



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from autologging import logged

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.junctions.junction import Junction


@logged
class NoSquid(Junction):
    """The PCell declaration for a No SQUID junction.

    Origin at the center of junction layer bottom edge.
    """

    junction_total_length = Param(pdt.TypeDouble, "Junction total length", 26.5, unit='µm')

    def build(self):
        # only add refpoints
        self.refpoints["origin_squid"] = pya.DPoint(0, 0)
        self.refpoints["port_common"] = pya.DPoint(0, self.junction_total_length)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_rf import FlipChipConnectorRf
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import add_parameters_from


@add_parameters_from(FlipChipConnectorRf)
@add_parameters_from(Simulation, face_stack=['1t1', '2b1'])
class FlipChipConnectorSim(Simulation):

    def build(self):
        fcc_cell = self.add_element(FlipChipConnectorRf, **self.get_parameters())

        transf = pya.DTrans(0, False, (self.box.left + self.box.right) / 2, (self.box.bottom + self.box.top) / 2)
        _, refp = self.insert_cell(fcc_cell, transf)

        self.produce_waveguide_to_port(refp["1t1_port"], refp["1t1_port_corner"], 1, 'left', face=0)
        diff_to_rotation = lambda x: abs(x - (self.output_rotation % 360))
        port_dir = {0: 'left', 90: 'bottom', 180: 'right', 270: 'top', 360: 'left'}\
            .get(min([0, 90, 180, 270, 360], key=diff_to_rotation))
        self.produce_waveguide_to_port(refp["2b1_port"], refp["2b1_port_corner"], 2, port_dir, face=1)



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import ast
import json
import os
import subprocess
from pathlib import Path
from typing import Callable, List, Tuple, Union
from kqcircuits.defaults import STARTUPINFO, XSECTION_PROCESS_PATH
from kqcircuits.pya_resolver import pya, klayout_executable_command
from kqcircuits.simulations.export.util import export_layers
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.simulations.simulation import Simulation


def xsection_call(input_oas: Path, output_oas: Path, cut1: pya.DPoint, cut2: pya.DPoint,
        process_path: Path = XSECTION_PROCESS_PATH, parameters_path: Path = None) -> None:
    """Calls on KLayout to run the XSection plugin

    Args:
        input_oas: Input OAS file (top-down geometry)
        output_oas: Output OAS file (Cross-section of input geometry)
        cut1: DPoint of first endpoint of the cross-section cut
        cut2: DPoint of second endpoint of the cross-section cut
        process_path: XSection process file that defines cross-section etching depths etc
        parameters_path: If process_path points to kqc_process.xs,
            parameters_path should point to the XSection parameters json file
            containing sweeped parameters and layer information.
    """
    if os.name == "nt":
        klayout_dir_name = "KLayout"
    elif os.name == "posix":
        klayout_dir_name = ".klayout"
    else:
        raise SystemError("Error: unsupported operating system")
    xsection_plugin_path = os.path.join(os.path.expanduser("~"), klayout_dir_name, "salt/xsection/macros/xsection.lym")
    cut_string = f"{cut1.x},{cut1.y};{cut2.x},{cut2.y}"

    if not klayout_executable_command():
        raise Exception("Can't find klayout executable command!")
    if not Path(xsection_plugin_path).is_file():
        raise Exception("The 'xsection' plugin is missing in KLayout! Go to 'Tools->Manage Packages' to install it.")

    # Hack: Weird prefix keeps getting added when path is converted to string which breaks the ruby plugin
    xs_run = str(process_path).replace("\\\\?\\", "")
    xs_params = str(parameters_path).replace("\\\\?\\", "")
    # When debugging, remove '-z' argument to see ruby error messages
    subprocess.run([klayout_executable_command(), input_oas.absolute(), '-z', '-nc', '-rx',
                    '-r', xsection_plugin_path,
                    '-rd', f'xs_run={xs_run}',
                    '-rd', f'xs_params={xs_params}',
                    '-rd', f'xs_cut={cut_string}',
                    '-rd', f'xs_out={output_oas.absolute()}'],
        check=True, startupinfo=STARTUPINFO)


# pylint: disable=dangerous-default-value
def create_xsections_from_simulations(simulations: List[Simulation],
                                      output_path: Path,
                                      cuts: Union[Tuple[pya.DPoint, pya.DPoint], List[Tuple[pya.DPoint, pya.DPoint]]],
                                      process_path: Path = XSECTION_PROCESS_PATH,
                                      ma_permittivity: float = 0,
                                      ms_permittivity: float = 0,
                                      sa_permittivity: float = 0,
                                      ma_thickness: float = 0,
                                      ms_thickness: float = 0,
                                      sa_thickness: float = 0,
                                      london_penetration_depth: float = 0,
                                      magnification_order: int = 0
                                    ) -> List[Simulation]:
    """Create cross-sections of all simulation geometries in the list.
    Will set 'box' and 'cell' parameters according to the produced cross-section geometry data.

    Args:
        simulations: List of Simulation objects, usually produced by a sweep
        output_path: Path for the exported simulation files
        cuts: 1. A tuple (p1, p2), where p1 and p2 are endpoints of a cross-section cut or
              2. a list of such tuples such that each Simulation object gets an individual cut
        process_path: XSection process file that defines cross-section etching depths etc
        ma_permittivity: Permittivity of metal–vacuum (air) interface
        ms_permittivity: Permittivity of metal–substrate interface
        sa_permittivity: Permittivity of substrate–vacuum (air) interface
        ma_thickness: Thickness of metal–vacuum (air) interface
        ms_thickness: Thickness of metal–substrate interface
        sa_thickness: Thickness of substrate–vacuum (air) interface
        london_penetration_depth: London penetration depth of the superconducting material
        magnification_order: Increase magnification of simulation geometry to accomodate more precise spacial units.
            0 =   no magnification with 1e-3 dbu
            1 =  10x magnification with 1e-4 dbu
            2 = 100x magnification with 1e-5 dbu etc
            Consider setting non-zero value when using oxide layers with < 1e-3 layer thickness or
            taking cross-sections of thin objects

    Returns:
        List of CrossSectionSimulation objects for each Simulation object in simulations
    """
    if isinstance(cuts, Tuple):
        cuts = [cuts] * len(simulations)
    if len(simulations) != len(cuts):
        raise Exception("Number of cuts did not match the number of simulations")
    if any(len(simulation.get_parameters()["face_stack"]) not in (1, 2) for simulation in simulations):
        raise Exception("Only single face and flip chip cross section simulations currently supported")

    xsection_dir = output_path.joinpath("xsection_tmp")
    xsection_dir.mkdir(parents=True, exist_ok=True)

    layout = pya.Layout()
    load_opts = _load_layout_options_for_xsection_output()
    for simulation, cut in zip(simulations, cuts):
        xsection_parameters = _dump_xsection_parameters(xsection_dir, simulation)
        simulation_file = xsection_dir / f"original_{simulation.cell.name}.oas"
        xsection_file   = xsection_dir / f"xsection_{simulation.cell.name}.oas"
        export_layers(str(simulation_file), simulation.layout, [simulation.cell],
                    output_format='OASIS',
                    layers=None)
        xsection_call(simulation_file, xsection_file, cut[0], cut[1], process_path, xsection_parameters)

        layout.read(str(xsection_file), load_opts)
        xsection_cell = layout.top_cells()[-1]
        xsection_cell.name = simulation.cell.name

    _clean_tmp_xsection_directory(xsection_dir, simulations)
    # Collect cross section simulation sweeps
    return [_construct_cross_section_simulation(
                layout,
                xsection_cell,
                simulations[idx],
                ma_permittivity,
                ms_permittivity,
                sa_permittivity,
                ma_thickness,
                ms_thickness,
                sa_thickness,
                london_penetration_depth,
                magnification_order)
        for idx, xsection_cell in enumerate(layout.top_cells())]
# pylint: enable=dangerous-default-value


def separate_signal_layer_shapes(simulation: Simulation, sort_key: Callable[[pya.Shape], float] = None):
    """Separate shapes in signal layer to their own dedicated signal layers for each face

    Args:
        simulation: A Simulation object where the layer will be separated
        sort_key: A function that, given a Shape object, returns a number.
            Shapes are sorted according to the number in increasing order.
            If None, picks a point in shape polygon, sorts points top to bottom then tie-breaks left to right
    """
    if sort_key is None:
        def sort_key(shape):
            point_in_shape = list(shape.polygon.each_point_hull())[0]
            return (-point_in_shape.y, point_in_shape.x)
    signal_index = 1
    gen_free_layer_slots = free_layer_slots(simulation.layout)
    for face in simulation.face_ids:
        signal_layer = find_layer_by_name(f"{face}_signal", simulation.layout)
        if signal_layer is None:
            continue
        signal_layer_idx = simulation.layout.layer(signal_layer)
        for shape in sorted(simulation.cell.each_shape(signal_layer_idx), key=sort_key):
            # Reuse layer if it already used in layout
            signal_layer = find_layer_by_name(f"{face}_signal_{signal_index}", simulation.layout)
            # If no such layer, find next available layer index
            if signal_layer is None:
                layer_index = next(gen_free_layer_slots)
                signal_layer = pya.LayerInfo(layer_index, 0, f"{face}_signal_{signal_index}")
            simulation.cell.shapes(simulation.layout.layer(signal_layer)).insert(shape)
            signal_index += 1
        simulation.cell.clear(signal_layer_idx)


def find_layer_by_name(layer_name, layout):
    """Returns layerinfo if there already is a layer by layer_name in layout. None if no such layer exists"""
    for l in layout.layer_infos():
        if l.datatype == 0 and layer_name == l.name:
            return l
    return None


def free_layer_slots(layout):
    """A generator of available layer slots"""
    layer_index = 0
    reserved_layer_ids = [l.layer for l in layout.layer_infos() if l.datatype == 0]
    while True:
        layer_index += 1
        if layer_index in reserved_layer_ids:
            continue
        yield layer_index


def _load_layout_options_for_xsection_output():
    load_opts = pya.LoadLayoutOptions()
    load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
    return load_opts


def _remap_face(layer_name, faces_of_flipchip):
    """Rename face to b_*, t_* convention based on faces_of_flipchip list"""
    if len(faces_of_flipchip) > 0 and layer_name.startswith(f"{faces_of_flipchip[0]}_"):
        return f"b_{layer_name[len(faces_of_flipchip[0]) + 1:]}"
    if len(faces_of_flipchip) > 1 and layer_name.startswith(f"{faces_of_flipchip[1]}_"):
        return f"t_{layer_name[len(faces_of_flipchip[1]) + 1:]}"
    return layer_name


def _dump_xsection_parameters(xsection_dir, simulation):
    """If we're sweeping xsection specific parameters,
    dump them in external file for xsection process file to pick up
    """
    simulation_params = {param_name: param_value for param_name, param_value in simulation.get_parameters().items()
                            if not isinstance(param_value, pya.DBox)} # Hack: ignore non-serializable params
    # Also dump all used layers in the simulation cell
    sim_layers = {_remap_face(l.name, simulation_params['face_stack']): f"{l.layer}/{l.datatype}"
        for l in simulation.layout.layer_infos()}
    # Find avaiable layer numbers for substrate layers
    gen_free_layer_slots = free_layer_slots(simulation.layout)
    sim_layers["b_substrate"] = f"{next(gen_free_layer_slots)}/0"
    sim_layers["t_substrate"] = f"{next(gen_free_layer_slots)}/0"
    simulation_params['sim_layers'] = sim_layers
    xsection_parameters_file = xsection_dir / f"parameters_{simulation.cell.name}.json"
    with open(xsection_parameters_file, "w") as sweep_file:
        json.dump(simulation_params, sweep_file)
    return xsection_parameters_file


def _clean_tmp_xsection_directory(xsection_dir, simulations):
    for simulation in simulations:
        if os.path.exists(xsection_dir / f"original_{simulation.cell.name}.oas"):
            os.remove(xsection_dir / f"original_{simulation.cell.name}.oas")
        if os.path.exists(xsection_dir / f"xsection_{simulation.cell.name}.oas"):
            os.remove(xsection_dir / f"xsection_{simulation.cell.name}.oas")
        if os.path.exists(xsection_dir / f"parameters_{simulation.cell.name}.json"):
            os.remove(xsection_dir / f"parameters_{simulation.cell.name}.json")
    if os.path.exists(xsection_dir):
        os.rmdir(xsection_dir)


def _combine_region_from_layers(simulation, layers):
    """Produce a region combined from regions in layers list"""
    region = pya.Region()
    for layer in layers:
        region += pya.Region(simulation.cell.shapes(simulation.layout.layer(layer)))
    return region


def _edge_on_the_box_border(edge, box):
    """True if edge is exactly at the rim of the box"""
    return  (edge.x1 == box.p1.x and edge.x2 == box.p1.x) or \
            (edge.x1 == box.p2.x and edge.x2 == box.p2.x) or \
            (edge.y1 == box.p1.y and edge.y2 == box.p1.y) or \
            (edge.y1 == box.p2.y and edge.y2 == box.p2.y)


def _cut_edge(target_edge, source_edge, extra_edges):
    """Cut an end of the target_edge with source_edge.

    If source_edge leaves behind two ends of the target_edge,
    the second edge bit is stored in extra_edges.
    """
    # Copy target_edge to not modify the original edge instance
    result_edge = pya.DEdge(target_edge.p1.x, target_edge.p1.y, target_edge.p2.x, target_edge.p2.y)
    if result_edge.contains_excl(source_edge.p1):
        if result_edge.contains_excl(source_edge.p2) and source_edge.p2 != result_edge.p2:
            extra_edges.append(pya.DEdge(source_edge.p2, result_edge.p2))
        result_edge.p2 = source_edge.p1
    elif result_edge.contains_excl(source_edge.p2):
        result_edge.p1 = source_edge.p2
    return result_edge


def _remove_shared_points(target_edge, acting_edges, is_adjacent):
    """Remove all points shared by target_edge and edges in acting_edges

    Returns a set of continuous edges that are not contained by acting_edges.
    Set is_adjacent to True if the shape of acting_edges is adjacent to the shape
    from which target_edge was taken. Set to False if the shapes are on top of eah other.
    """
    edge_bits = [target_edge]
    for acting_edge in acting_edges:
        # Set acting_edge to point to same direction as target_edge
        if is_adjacent:
            acting_edge = acting_edge.swapped_points()
        # Consider edges if they share points, which means they are parallel and have same displacement
        if acting_edge.is_parallel(target_edge):
            # Remove edge bits if they are completely covered by acting_edge
            edge_bits = [e for e in edge_bits if not (acting_edge.contains(e.p1) and acting_edge.contains(e.p2))]
            extra_edge_bits = [] # Collect extra edge bits here
            edge_bits = [_cut_edge(e, acting_edge, extra_edge_bits) for e in edge_bits]
            edge_bits.extend(extra_edge_bits) # Add extra bits
            edge_bits = [e for e in edge_bits if e.p1 != e.p2] # Remove zero length edge bits
    return edge_bits


def _thicken_edges(edges, thickness, dbu, grow):
    """Take edges and add thickness to produce a region.

    Requires dbu.
    Set grow to True to grow the region outward, False to grow inward
    """
    if thickness <= 0.0: # Don't do anything if no thickness
        return pya.Region()
    # Construct a graph from the edges to find paths
    # Start by finding start points for paths
    start_points = [e.p1 for e in edges if e.p1 not in [e2.p2 for e2 in edges]]
    path_graph = {}
    for edge in edges:
        path_graph[edge.p1] = edge

    result_layer = pya.Region()
    # Take each start_point and follow the path until the end
    for current_point in start_points:
        polygon_points = [current_point]
        normals = []
        while True:
            # First collect path points for the region polygon
            polygon_points.append(path_graph[current_point].p2)
            edge_dir = path_graph[current_point].p2 - path_graph[current_point].p1
            # Store edge normal, assuming edges go clock-wise around the shape hull
            normal = pya.DPoint(-edge_dir.y, edge_dir.x)
            if not grow: # Flip normal if growing inward
                normal = -normal
            # Set normal length to thickness
            normals.append(normal * (thickness / normal.abs()))
            # At the end point, terminate
            if path_graph[current_point].p2 not in path_graph:
                break
            # Otherwise proceed to next point in path
            current_point = path_graph[current_point].p2
        # Connect to the second layer of the path to add thickness
        polygon_points.append(polygon_points[-1] + normals[-1])
        # Backtrack the path for the second layer of the polygon
        for idx in range(len(normals) - 1, 0, -1):
            normal_sum = normals[idx] + normals[idx - 1] # Sum normals of surrounding edges of the point
            polygon_points.append(polygon_points[idx] + normal_sum)
        polygon_points.append(polygon_points[0] + normals[0]) # Last second layer point, copied from the start_point
        result_layer += pya.Region(pya.DPolygon(polygon_points).to_itype(dbu))
    return result_layer


def _oxidise_layers(simulation, ma_thickness, ms_thickness, sa_thickness):
    """Take the cross section geometry and add oxide layers between substrate, metal and vaccuum.
    Will add geometry around metals and etch away substrate to insert oxide geometry.
    """
    substrate_layers = [layer for layer in simulation.layout.layer_infos() if layer.name.endswith("_substrate")]
    substrate = _combine_region_from_layers(simulation, substrate_layers)
    metal_layers  = [layer for layer in simulation.layout.layer_infos() if layer.name in
        ["b_ground", "t_ground", "b_signal", "t_signal"]]
    metal_layers += [layer for layer in simulation.layout.layer_infos() if layer.name.startswith("b_signal_")]
    metal_layers += [layer for layer in simulation.layout.layer_infos() if layer.name.startswith("t_signal_")]
    metals = _combine_region_from_layers(simulation, metal_layers)
    metal_edges = [e.to_dtype(simulation.layout.dbu) for e in metals.edges()]
    substrate_edges = [e.to_dtype(simulation.layout.dbu) for e in substrate.edges()]

    ma_edges = []
    for metal_edge in metal_edges:
        if not _edge_on_the_box_border(metal_edge, simulation.box):
            ma_edges.extend(_remove_shared_points(metal_edge, substrate_edges, True))

    sa_edges, ms_edges = [], []
    for substrate_edge in substrate_edges:
        if not _edge_on_the_box_border(substrate_edge, simulation.box):
            sa_edges.extend(_remove_shared_points(substrate_edge, metal_edges, True))
            ms_edges.extend(_remove_shared_points(substrate_edge, sa_edges, False))

    ma_layer = _thicken_edges(ma_edges, ma_thickness, simulation.layout.dbu, True)
    ms_layer = _thicken_edges(ms_edges, ms_thickness, simulation.layout.dbu, False)
    sa_layer = _thicken_edges(sa_edges, sa_thickness, simulation.layout.dbu, True)
    sa_layer -= ma_layer # MA layer takes precedence over SA layer

    # Etch and replace substrate layer regions
    if ms_thickness > 0.0 or sa_thickness > 0.0:
        for substrate_layer in substrate_layers:
            substrate_region = pya.Region(simulation.cell.shapes(simulation.layout.layer(substrate_layer)))
            simulation.cell.shapes(simulation.layout.layer(substrate_layer)).clear()
            simulation.cell.shapes(simulation.layout.layer(substrate_layer)).insert(
                substrate_region - ms_layer)

    if ma_thickness > 0.0:
        simulation.cell.shapes(simulation.get_sim_layer("ma_layer")).insert(ma_layer)
    if ms_thickness > 0.0:
        simulation.cell.shapes(simulation.get_sim_layer("ms_layer")).insert(ms_layer)
    if sa_thickness > 0.0:
        simulation.cell.shapes(simulation.get_sim_layer("sa_layer")).insert(sa_layer)


def _construct_cross_section_simulation(layout, xsection_cell, simulation,
        ma_permittivity, ms_permittivity, sa_permittivity,
        ma_thickness, ms_thickness, sa_thickness,
        london_penetration_depth, magnification_order):
    """Produce CrossSectionSimulation object"""
    if magnification_order > 0:
        layout.dbu = 10 ** (-3 - magnification_order)
        xsection_cell.transform(pya.DCplxTrans(10 ** magnification_order))
    xsection_parameters = simulation.get_parameters()
    xsection_parameters['london_penetration_depth'] = london_penetration_depth
    cell_bbox = xsection_cell.dbbox()
    # Disabled for single face and flip-chip cases
    #cell_bbox.p1 -= pya.DPoint(0, xsection_parameters['lower_box_height'])
    if len(xsection_parameters['face_stack']) == 1:
        cell_bbox.p2 += pya.DPoint(0, xsection_parameters['upper_box_height'])
    xsection_parameters['box'] = cell_bbox
    xsection_parameters['cell'] = xsection_cell
    xsection_simulation = CrossSectionSimulation(layout, **xsection_parameters)
    # Keep all parameters given in simulations for JSON
    for k, v in xsection_parameters.items():
        setattr(xsection_simulation, k, v)
    xsection_simulation.xsection_source_class = type(simulation)
    xsection_simulation.register_cell_layers_as_sim_layers()

    material_dict = xsection_parameters['material_dict']
    material_dict = ast.literal_eval(material_dict) if isinstance(material_dict, str) else material_dict
    substrate_material = xsection_parameters['substrate_material']
    b_substrate_permittivity = material_dict[substrate_material[0]]['permittivity']
    xsection_simulation.set_permittivity('b_substrate', b_substrate_permittivity)
    if len(xsection_parameters['face_stack']) == 2:
        t_substrate_permittivity = b_substrate_permittivity
        if len(substrate_material) > 1:
            t_substrate_permittivity = material_dict[substrate_material[1]]['permittivity']
        xsection_simulation.set_permittivity('t_substrate', t_substrate_permittivity)
    _oxidise_layers(xsection_simulation, ma_thickness, ms_thickness, sa_thickness)
    if ma_thickness > 0.0:
        xsection_simulation.set_permittivity('ma_layer', ma_permittivity)
    if ms_thickness > 0.0:
        xsection_simulation.set_permittivity('ms_layer', ms_permittivity)
    if sa_thickness > 0.0:
        xsection_simulation.set_permittivity('sa_layer', sa_permittivity)
    return xsection_simulation



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import copy
import os
import logging
from sys import argv
from time import perf_counter
from inspect import isclass
from multiprocessing import Pool
from pathlib import Path

from autologging import logged
from tqdm import tqdm

from kqcircuits.util.log_router import route_log
from kqcircuits.pya_resolver import pya, is_standalone_session
from kqcircuits.defaults import default_bar_format, TMP_PATH, default_face_id
from kqcircuits.masks.mask_export import export_chip, export_mask_set
from kqcircuits.masks.mask_layout import MaskLayout
from kqcircuits.klayout_view import KLayoutView


@logged
class MaskSet:
    """Class representing a set of masks for different chip faces.

    A mask set consists of one or more MaskLayouts, each of which is for a certain face.

    To create a mask, add mask layouts to the mask set using add_mask_layout() and add chips to these mask layouts using
    add_chip(). These functions also export some files for each chip. Then call build() to create the
    cell hierarchy of the entire mask, and finally export mask files by calling export().

    Example:
        mask = MaskSet(...)
        mask.add_mask_layout(...)
        mask.add_mask_layout(...)
        mask.add_chip(...)
        mask.build()
        mask.export()

    Attributes:
        layout: pya.Layout of this mask set
        name: Name of the mask set
        version: Version of the mask set
        with_grid: Boolean determining if ground grid is generated
        export_drc: Boolean determining if DRC report is exported
        chips_map_legend: Dictionary where keys are chip names, values are chip cells
        mask_layouts: list of MaskLayout objects in this mask set
        mask_export_layers: list of names of the layers which are exported for each MaskLayout
        used_chips: similar to chips_map_legend, but only includes chips which are actually used in mask layouts
        export_path: The folder for mask files will be generated under this. TMP_PATH by default.
    """

    def __init__(self, view=None, name="MaskSet", version=1, with_grid=False, export_drc=False,
                 mask_export_layers=None, export_path=TMP_PATH):

        self._time = {"INIT": perf_counter(), "ADD_CHIPS": 0,  "BUILD": 0, 'EXPORT': 0, 'END': 0}

        if view is None:
            self.view = KLayoutView()
        self.layout = self.view.layout

        self.name = name
        self.version = version
        self.with_grid = with_grid
        self.export_drc = export_drc
        self.chips_map_legend = {}
        self.mask_layouts = []
        self.mask_export_layers = mask_export_layers if mask_export_layers is not None else []
        self.used_chips = {}
        self._extra_params = {}
        self._mask_set_dir = Path(export_path)/f"{name}_v{version}"

        self._mask_set_dir.mkdir(parents=True, exist_ok=True)

        self._extra_params["enable_debug"] = '-d' in argv
        self._single_process = self._extra_params["enable_debug"] or not is_standalone_session()

        self._cpu_override = 0
        if '-c' in argv and len(argv) > argv.index('-c') + 1:
            self._cpu_override = int(argv[argv.index('-c') + 1])


    def add_mask_layout(self, chips_map, face_id=default_face_id, mask_layout_type=MaskLayout, **kwargs):
        """Creates a mask layout from chips_map and adds it to self.mask_layouts.

        Args:
            chips_map: List of lists (2D-array) of strings, each string is a chip name (or --- for no chip)
            face_id: face_id of the mask layout
            mask_layout_type: type of the mask layout (MaskLayout or a child class of it)
            kwargs: keyword arguments passed to the mask layout

        Returns:
            the created mask layout
        """
        if ("mask_export_layers" not in kwargs) and self.mask_export_layers:
            kwargs["mask_export_layers"] = self.mask_export_layers

        mask_layout = mask_layout_type(self.layout, self.name, self.version, self.with_grid, chips_map, face_id,
                                       **kwargs)
        self.mask_layouts.append(mask_layout)
        return mask_layout

    def add_chip(self, chips, variant_name=None, cpus=None, **kwargs):
        """Adds a chip (or list of chips) with parameters to self.chips_map_legend and exports the files for each chip.

        Note the complex polymorphism used here: ``chips`` is either a single chip class or a list of ``(chip, variant,
        parameters)`` tuples. In the latter case the rest of the arguments (except ``cpus``) are ignored. Also,
        ``chips`` (or the individual chip part of tuples) may be a simple file name to load a static .oas file instead.

        Chips are created in parallel in separate processes but the user may choose to use a ``-d`` switch on
        the command line for debugging with a single process. It is also possible to manually limit the number of
        concurrently used CPUs for resource management purposes with the ``-c 4`` switch (to 4 in this example).

        Args:
            chip: A chip class. Or a list of tuples, like ``[(QualityFactor, "QDG", parameters),...]``,
                  parameters are optional.
            variant_name: Name for specific variant, the same as in the mask layout. Or None, if multiple
                          chips are given
            cpus: Number of parallel processes to use for chip generation. By default uses ``os.cpu_count()``
                  or the number of chips, whichever is smaller.
            **kwargs: Any parameters passed to the a single chip PCell. Not used with multiple chips.
        """
        self._time['ADD_CHIPS'] = perf_counter()

        if not isinstance(chips, list):  # only one chip
            cpus = 1
            chips = [(chips, variant_name, kwargs)]

        if cpus is None:
            cpus = min(len(chips), os.cpu_count())
        if self._cpu_override > 0:
            cpus = self._cpu_override

        # Pool.map() needs all arguments packed into a single list
        xargs = (self.name, self.with_grid, self._mask_set_dir, self.export_drc, self._extra_params)
        chip_args = ((chip, xargs) for chip in chips)

        file_names = []
        if cpus == 1 or self._single_process:
            file_names += map(self._create_chip, chip_args)
        else:
            print(f"Building chip variants in parallel using {cpus} processes...")
            with Pool(cpus) as pool:
                file_names += pool.map(self._create_chip, chip_args)

        # import chip cells exported by the parallel processes into the mask
        for variant, file_name in tqdm(file_names, desc='Add chips into mask', bar_format=default_bar_format):
            self._load_chip_into_mask(file_name, variant)

    @staticmethod
    def _create_chip(chip_arg):
        """Create chip, possibly in a separate process."""

        chip, xargs = chip_arg
        chip_class, variant_name, *chip_params = chip
        name, with_grid, _mask_set_dir, export_drc, _extra_params = xargs

        chip_path = _mask_set_dir/"Chips"/f"{variant_name}"
        chip_path.mkdir(parents=True, exist_ok=True)

        logging.basicConfig(level=logging.DEBUG)  # this level is NOT actually used
        route_log(filename=chip_path/f"{variant_name}.log", stdout=_extra_params["enable_debug"])

        view = KLayoutView()
        layout = view.layout

        if isclass(chip_class):
            params = {
                'name_chip': variant_name,
                'name_mask': name,
                'with_grid': with_grid,
                'merge_base_metal_gap': True,
                'display_name': variant_name,
                'name_copy': None,
            }
            if chip_params:
                params.update(chip_params[0])
            cell = chip_class.create(layout, **params)
        else:  # its a file name, load it
            load_opts = pya.LoadLayoutOptions()
            if hasattr(pya.LoadLayoutOptions, "CellConflictResolution"):
                load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
            layout.read(chip_class, load_opts)
            cell = layout.top_cells()[-1]

        export_chip(cell, variant_name, chip_path, layout, export_drc)
        view.close()

        return variant_name, str(chip_path / f"{variant_name}.oas")

    def build(self, remove_guiding_shapes=True):
        """Builds the mask set.

        Creates cells for the mask based on self.mask_layouts and self.chips_map_legend. Optionally removes
        guiding shapes from the layout. Populates self.used_chips with the chips used in the mask layouts.

        Args:
            remove_guiding_shapes (Boolean): determines if the guiding shapes are removed

        """
        self._time['BUILD'] = perf_counter()
        # build mask layouts (without chip copy labels)
        for mask_layout in self.mask_layouts:
            # include face_id in mask_layout.name only for multi-face masks
            if len(self.mask_layouts) > 1:
                mask_layout.name += "-" + mask_layout.face_id
            mask_layout.build(self.chips_map_legend)

        chip_copy_label_layers = [
            "base_metal_gap",
            "base_metal_gap_wo_grid",
            "base_metal_gap_for_EBL"
        ]

        # Insert submask cells to different cell instances, so that these cells can have different chip labels even if
        # the original submask cells are identical. Also copy the MaskLayout objects of identical submasks into separate
        # MaskLayout objects with different `extra_id` so that mask export can use that information.
        mask_layouts_to_remove = set()
        submask_layouts_with_exported_layers = set()
        submask_layouts = []
        for mask_layout in self.mask_layouts:
            for i, (sm_layout, sm_pos) in enumerate(mask_layout.submasks):
                new_sm_layout = copy.copy(sm_layout)
                new_sm_layout.extra_id = f" s{i+1}"
                old_top_cell = new_sm_layout.top_cell
                new_sm_layout.top_cell = self.layout.create_cell(f"{new_sm_layout.name}{new_sm_layout.extra_id}")
                new_sm_layout.top_cell.insert(pya.DCellInstArray(old_top_cell.cell_index(), pya.DTrans()))
                mask_layout.top_cell.insert(
                    pya.DCellInstArray(new_sm_layout.top_cell.cell_index(),
                                       pya.DTrans(sm_pos - sm_layout.wafer_center + mask_layout.wafer_center))
                )
                mask_layout.submasks[i] = (new_sm_layout, sm_pos)
                submask_layouts.append(new_sm_layout)
                mask_layouts_to_remove.add(sm_layout)
                # Make sure that layers are only exported once if there are multiple identical submasks
                if sm_layout in submask_layouts_with_exported_layers:
                    # only export layers where chip copy labels are since they are different even for identical submasks
                    new_sm_layout.mask_export_layers = \
                        [layer for layer in chip_copy_label_layers if layer in new_sm_layout.mask_export_layers]
                else:
                    submask_layouts_with_exported_layers.add(sm_layout)
        self.mask_layouts = submask_layouts + [ml for ml in self.mask_layouts if ml not in mask_layouts_to_remove]

        # add chip copy labels for every mask layout
        for mask_layout in tqdm(self.mask_layouts, desc='Adding chip copy labels', bar_format=default_bar_format):

            labels_cell = mask_layout.layout.create_cell("ChipLabels")
            mask_layout.top_cell.insert(pya.DCellInstArray(labels_cell.cell_index(), pya.DTrans(pya.DVector(0, 0))))

            if mask_layout not in submask_layouts:
                mask_layout.insert_chip_copy_labels(labels_cell, chip_copy_label_layers)
                # remove "$1" or similar unnecessary postfix from cell name
                mask_layout.top_cell.name = f"{mask_layout.name}"

        # populate used_chips with chips which exist in some mask_layout
        for chip_name, cell in self.chips_map_legend.items():
            for mask_layout in self.mask_layouts:
                # pylint: disable=use-a-generator
                if any([chip_name in row for row in mask_layout.chips_map]):
                # pylint: enable=use-a-generator
                    self.used_chips[chip_name] = cell
                    break

        # remove the guiding shapes, like chip boxes and waveguide paths
        if remove_guiding_shapes and self.layout.is_valid_layer(self.layout.guiding_shape_layer()):
            self.layout.delete_layer(self.layout.guiding_shape_layer())

    def export(self):
        """Exports designs, bitmaps and documentation of this mask set.

        Assumes that self.build() has been called before.
        """
        self._time['EXPORT'] = perf_counter()

        print("Exporting mask set...")
        export_mask_set(self)

        self._time['END'] = perf_counter()

        def tdiff(a, b):  # get elapsed time from "a" to "b"
            return f'{self._time[b] - self._time[a]:.1f}s' if self._time[a] and self._time[b] else 'n/a'

        print(f"Runtime: {tdiff('INIT', 'END')} (add chips: {tdiff('ADD_CHIPS', 'BUILD')}, "
              f"build: {tdiff('BUILD', 'EXPORT')}, export: {tdiff('EXPORT', 'END')})")

    @staticmethod
    def chips_map_from_box_map(box_map, mask_map):
        """Returns the chips_map created from box_map and mask_map.

        Given NxN box map and MxM mask_map, creates chips_map of size MNxMN. So each element of mask map is "replaced"
        by a box in the box map. Assumes that box_map and mask_map are square.

        Args:
            box_map: dictionary where keys are strings identifying the box type, and values are 2D arrays (lists of
                lists) where each element is a string identifying the chip type
            mask_map: 2D array (list of lists), where each element is a string identifying the box type

        """
        num_box_map_rows = len(list(box_map.values())[0])
        num_mask_map_rows = len(mask_map)
        num_chip_rows = num_box_map_rows*num_mask_map_rows

        chips_map = [["" for _ in range(num_chip_rows)] for _ in range(num_chip_rows)]
        for (k, box_row) in enumerate(mask_map):
            for (l, box) in enumerate(box_row):
                if box in box_map:
                    for (i, row) in enumerate(box_map[box]):
                        for (j, slot) in enumerate(row):
                            chips_map[k*num_box_map_rows + i][l*num_box_map_rows + j] = slot

        return chips_map

    def _load_chip_into_mask(self, file_name, variant_name):
        """Loads a chip from file_name to self.layout and adds it into self.chips_map_legend["variant_name"]"""
        load_opts = pya.LoadLayoutOptions()
        load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
        self.layout.read(file_name, load_opts)
        self.chips_map_legend.update({variant_name: self.layout.top_cells()[-1]})



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import ast
from itertools import zip_longest
from typing import Tuple
from math import pi, tan
from autologging import logged

from scipy.optimize import root_scalar

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.util.library_helper import element_by_class_name
from kqcircuits.util.geometry_helper import vector_length_and_direction, point_shift_along_vector, \
    get_cell_path_length, get_angle, get_direction
from kqcircuits.elements.element import Element
from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.airbridge_connection import AirbridgeConnection
from kqcircuits.elements.meander import Meander
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_taper import WaveguideCoplanarTaper
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_rf import FlipChipConnectorRf


class Node:
    """Specifies a single node of a composite waveguide.

    Node is as a ``position`` and optionally other parameters. The ``element`` argument sets an Element
    type that gets inserted in the waveguide. Typically this is an Airbridge, but any element with
    port_a and port_b is supported.

    Args:
        position: The location of the Node. Represented as a DPoint object.
        element: The Element type that gets inserted in the waveguide. None by default.
        inst_name: If an instance name is supplied, the element refpoints will be exposed with that name. Default None.
        align: Tuple with two refpoint names that correspond the input and output point of element, respectively
               Default value (None) uses ``('port_a', 'port_b')``
        angle: Angle of waveguide direction in degrees
        length_before: Length of the waveguide segment before this node
        length_increment: Waveguide length increment produced by meander before this node
        **params: Other optional parameters for the inserted element

    Returns:
        A Node.

    .. MARKERS_FOR_PNG 180,1.2
    """

    position: pya.DPoint
    element: Element
    inst_name: str
    align: Tuple
    angle: float
    length_before: float
    length_increment: float

    def __init__(self, position, element=None, inst_name=None, align=tuple(), angle=None, length_before=None,
                 length_increment=None, **params):
        if isinstance(position, tuple):
            self.position = pya.DPoint(position[0], position[1])
        else:
            self.position = position
        self.element = element
        self.align = align
        self.inst_name = inst_name
        self.angle = angle
        self.length_before = length_before
        self.length_increment = length_increment
        self.params = params

    def __str__(self):
        """
        String representation of a Node, used for serialization and needed for storage in KLayout parameters.

        The corresponding deserialization is implemented in `Node.deserialize`.
        """

        txt = f"{self.position.x}, {self.position.y}"
        if self.element is not None:
            txt += f", '{self.element.__name__}'"

        magic_params = {}
        if self.align:
            magic_params['align'] = self.align
        if self.inst_name:
            magic_params['inst_name'] = self.inst_name
        if self.angle is not None:
            magic_params['angle'] = self.angle
        if self.length_before is not None:
            magic_params['length_before'] = self.length_before
        if self.length_increment is not None:
            magic_params['length_increment'] = self.length_increment

        all_params = {**self.params, **magic_params}
        if all_params:
            for pn, pv in all_params.items():
                if isinstance(pv, pya.DPoint):  # encode DPoint as tuple
                    all_params[pn] = (pv.x, pv.y)
            txt += f", {all_params}"
        return "(" + txt + ")"

    @classmethod
    def deserialize(cls, node):
        """
        Create a Node object from a serialized form, such that ``from_serialized(ast.literal_eval(str(node_object)))``
        returns an equivalent copy of ``node_obj``.

        Args:
            node: serialized node, consisting of a tuple ``(x, y, element_name, params)``, where ``x`` and ``y`` are the
                node coordinates. The string ``element_name`` and dict ``params`` are optional.

        Returns: a Node

        """
        x, y = float(node[0]), float(node[1])
        element = None
        params = {}
        if len(node) > 2:
            if isinstance(node[2], dict):
                params = node[2]
            else:
                element = node[2]
        if len(node) > 3:
            params = node[3]

        if element is not None:
            if element in globals():
                element = globals()[element]
            else:
                element = element_by_class_name(element)

        # re-create DPoint from tuple
        magic_params = ('align', 'inst_name', 'angle')
        for pn, pv in params.items():
            if isinstance(pv, tuple) and pn not in magic_params:
                if len(pv) < 2:
                    raise ValueError(f'Point parameter {pn} should have two elements')
                params[pn] = pya.DPoint(float(pv[0]), float(pv[1]))

        return cls(pya.DPoint(x, y), element, **params)

    @staticmethod
    def nodes_from_string(nodes):
        """Converts the human readable text representation of Nodes to an actual Node object list.

        Needed for storage in KLayout parameters. The string has to conform to a specific format:
        `(x, y, class_str, parameter_dict)`. For example `(0, 500, 'Airbridge', {'n_bridges': 2})`,
        see also the `Node.__str__` method. Empty class_str or parameter_dict may be omitted.

        Returns:
            list of Node objects
        """

        nlas = ", ".join(nodes) if isinstance(nodes, list) else nodes
        node_list = ast.literal_eval(nlas + ",")

        return [Node.deserialize(node) for node in node_list]


@add_parameters_from(WaveguideCoplanarTaper, taper_length=100)
@add_parameters_from(Airbridge, "airbridge_type")
@add_parameters_from(FlipChipConnectorRf)
@add_parameters_from(WaveguideCoplanar, "term1", "term2")
@logged
class WaveguideComposite(Element):
    """A composite waveguide made of waveguides and other elements.

    From the user's perspective this is a WaveguideCoplanar that is extended with Airbridge,
    WaveguideCoplanarTaper and FlipChipConnector, so that a signal can be routed over other
    waveguides and several chip faces. As a bonus feature other arbitrary elements like coplanar
    capacitors are also supported.

    A list of Nodes defines the shape of the waveguide. An empty node, i.e. ``Node.element`` is
    ``None``, will set the next "waypoint" of the waveguide. If an Element is given it will be
    placed on the node, except for the first or last node where it will be adjacent to the edge.

    Inserted elements are oriented collinear with the previous node. If the following node is not
    collinear with the previous element then for convenience an automatic bend is inserted in the
    waveguide. Note that back-to-back elements (i.e. without some empty Nodes in between) should
    only be placed collinear with each other.

    Node parameters will be passed to the element. For convenience, empty nodes may also have
    parameters: ``a``, ``b`` or ``face_id``. They insert a WaveguideCoplanarTaper or a
    FlipChipConnector, respectively and change the defaults too.

    The ``ab_across=True`` parameter places a single airbridge across the node. The ``n_bridges=N``
    parameter puts N airbridges evenly distributed across the preceding edge.

    The ``length_before`` parameter of a node can be specified to automatically set the length of
    the waveguide between that node and the previous one. It will in fact create a Meander element
    instead of a normal waveguide between those nodes to achieve the correct length. Alternative parameter
    ``length_increment`` sets the waveguide length increment compared to normal waveguide.

    A notable implementation detail is that every Airbridge (sub)class is done as AirbridgeConnection.
    This way a waveguide taper is automatically inserted before and after the airbridge so the user
    does not have to manually add these. Other Node types do not have this feature.

    A WaveguideComposite cell has a method ``segment_lengths`` that returns a list of lengths of each individual
    regular waveguide segment. Segments are bounded by any element that is not a standard waveguide, such as Airbridge,
    flip chip, taper or any custom element.

    For examples see the test_waveguide_composite.lym script.
    """

    nodes = Param(pdt.TypeString, "List of Nodes for the waveguide", "(0, 0, 'Airbridge'), (200, 0)")
    enable_gui_editing = Param(pdt.TypeBoolean, "Enable GUI editing of the waveguide path", True)
    gui_path = Param(pdt.TypeShape, "", pya.DPath([pya.DPoint(0, 0), pya.DPoint(200, 0)], 1))
    gui_path_shadow = Param(pdt.TypeShape, "Hidden path to detect GUI operations",
                            pya.DPath([pya.DPoint(0, 0), pya.DPoint(200, 0)], 1), hidden=True)
    tight_routing = Param(pdt.TypeBoolean, "Tight routing for corners", False)

    @classmethod
    def create(cls, layout, library=None, **parameters):
        # For code-generated cells, make sure the gui path matches te node definition.
        if parameters.get("enable_gui_editing", True):
            if "nodes" in parameters:
                path = pya.DPath([node.position for node in parameters["nodes"]], 1)
                # Round to database units since the KLayout partial tool also works in database units
                path = path.to_itype(layout.dbu).to_dtype(layout.dbu)
                parameters["gui_path"] = path
                parameters["gui_path_shadow"] = path
        else:
            parameters["gui_path"] = pya.DPath()
            parameters["gui_path_shadow"] = pya.DPath()

        cell = super().create(layout, library, **parameters)
        setattr(cell, "segment_lengths", lambda: WaveguideComposite.get_segment_lengths(cell))

        return cell

    @staticmethod
    def get_segment_cells(cell):
        """Get the subcells of a ``WaveguideComposite`` cell, ordered by node index.

        The subcells include ``WaveguideCoplanar``, ``WaveguideCoplanarTaper`` and ``Meander`` cells for straight,
        tapered and meandering segments, respectively, and any element cells that were inserted explicitly
        (``element`` argument) or implicitly (changing ``a``, ``b`` or ``face_id``) at any ``Node``.

        The ``node_index`` returned with each subcell is an index to the ``nodes`` parameter of the
        ``WaveguideComposite`` cell. It points to the node that _created_ the subcell, which is the node following a
        segment for regular ``WaveguideCoplanar`` segments, and the node that specified the element or parameter change
        otherwise.

        Note that there may be multiple subcells per node, and some nodes may not have associated subcells (in
        particular, regular ``WaveguideCoplanar`` segments are merged when possible). Subscells are returned in the
        order in which they appear in the waveguide.

        Args:
            cell: A WaveguideComposite cell. Can be a PCell or static cell created from a PCell.

        Returns: A list of tuples (node_index: int, subcell: pya.Cell) ordered by node_index.
        """
        layout = cell.layout()

        segment_data = []

        for inst in cell.each_inst():
            # Note: Using layout.cell(inst.cell_index) instead of inst.cell to work around KLayout issue #235
            child_cell = layout.cell(inst.cell_index)

            node_index = inst.property("waveguide_composite_node_index")
            if node_index is not None:
                segment_data.append((node_index, child_cell))

        return list(sorted(reversed(segment_data), key=lambda x: x[0]))

    @staticmethod
    def get_segment_lengths(cell):
        """ Retrieves the segment lengths of each waveguide segment in a WaveguideComposite cell.

        Waveguide segments are ``WaveguideCoplanar``, ``WaveguideCoplanarTaper`` and ``Meander`` subcells.
        This method returns a list with the same length as the ``nodes`` parameter, where each element is the total
        length of all waveguides directly preceding and/or defined in that node. For example, for a taper node both the
        preceding regular waveguide and the taper itself are counted.

        Note that ``WaveguideComposite`` merges consecutive waveguide segments if they have no special elements. As
        a consequence, the corresponding waveguide lengths all accumulate in the next index which has a taper, meander
        or other element, and the length for nodes that contain ony ``WaveguideCoplanar`` is reported as 0.

        Args:
            cell: A WaveguideComposite cell. Can be a PCell or static cell created from a PCell.

        Returns: A list waveguide lengths per node
        """

        # Measure segment lengths, counting only "regular waveguides"
        waveguide_segment_types = {"Waveguide Coplanar", "Waveguide Coplanar Taper", "Meander"}

        segment_data = WaveguideComposite.get_segment_cells(cell)
        if len(segment_data) == 0:
            return []

        segment_lengths = [0] * (segment_data[-1][0] + 1)
        for node_index, child_cell in segment_data:
            if child_cell.name.split('$')[0] in waveguide_segment_types:
                segment_lengths[node_index] += get_cell_path_length(child_cell)

        return segment_lengths

    @staticmethod
    def produce_fixed_length_waveguide(chip, route_function, initial_guess=0.0, length=0.0, **waveguide_params):
        """
        Produce a waveguide composite with fixed length. `route_function` should be a single-argument function that
        returns route, and its argument is an adjustable length in µm.
        Note that this is not a minimization, but a single-step adjustment that only corrects for the offset in
        length.

        Args:
            chip: Chip in which the element is added (self if called within chip code)
            route_function: a function lambda x: [Node(f(x))...] where x can be for instance a meander length or a
            DPoint coordinate (if more than one component is tuned in the Node list, the correction length must be
            weighted)
            initial_guess: float that allows to draw an initial waveguide of a reasonable length
            length: target desired length for the final waveguide
            waveguide_params: kwargs to be passed to the WaveguideComposite element, such as a, b, term1, term2 etc.

        Returns: The waveguide instance, refpoints and the final length
        """

        wg_tmp = chip.add_element(WaveguideComposite, nodes=[
            *route_function(initial_guess),
        ], **waveguide_params)
        offset_length = wg_tmp.length()
        correction = length - offset_length
        wg = chip.add_element(WaveguideComposite, nodes=[
            *route_function(correction+initial_guess),
        ], **waveguide_params)
        inst, ref = chip.insert_cell(wg)
        return inst, ref, wg.length()

    def snap_point(self, point: pya.DPoint) -> pya.DPoint:  # pylint: disable=no-self-use
        """
        Interface to define snap behavior for GUI editing in derived classes.

        Args:
            point: DPoint that has been moved in the GUI

        Returns: DPoint representing the snapped location for the input point. Return the point unmodified if no
            snapping is required
        """

        # Default implementation always returns
        return point

    def coerce_parameters_impl(self):
        if self.enable_gui_editing:
            nodes = Node.nodes_from_string(self.nodes)

            # If gui_path was edited by the user, it is now different from gui_path_shadow. Detect changes
            changed = self.gui_path != self.gui_path_shadow

            if changed:
                new_points = list(self.gui_path.each_point())
                old_points = list(self.gui_path_shadow.each_point())

                length_change = len(new_points) - len(old_points)
                if length_change == 0:
                    # One or more points were moved; update node positions of the moved points
                    for node, new_position, old_position in zip(nodes, new_points, old_points):
                        if new_position != old_position:
                            new_position = self.snap_point(new_position)
                            node.position = new_position
                    self.nodes = [str(node) for node in nodes]
                elif length_change == 1:
                    # One node was added. Figure out which node it was.
                    added_index = None
                    added_point = None
                    for i, (new_point, old_point) in enumerate(zip_longest(new_points, old_points)):
                        if old_point is None or new_point != old_point:
                            added_index = i
                            added_point = new_point
                            break

                    if added_index is not None:
                        nodes.insert(added_index, Node(added_point))
                        self.nodes = [str(node) for node in nodes]
                elif length_change < 0 and self.gui_path.num_points() >= 2:
                    # One or more points deleted; delete corresponding nodes. We require at least two remaining nodes.
                    remaining_indices = []
                    new_indices = []

                    for i, old_point in enumerate(old_points):
                        if old_point in new_points:
                            remaining_indices.append(i)
                            new_indices.append(new_points.index(old_point))

                    # Verify we found the right number of remaining indices, and the order is correct
                    new_indices_monotonic = all(i < j for i, j in zip(new_indices, new_indices[1:]))
                    if len(remaining_indices) == len(new_points) and new_indices_monotonic:
                        nodes = [nodes[i] for i in remaining_indices]
                        self.nodes = [str(node) for node in nodes]

            # After coerce the gui_path and gui_path_shadow should both match the nodes.
            new_path = pya.DPath([node.position for node in nodes], 1)
            # Force rounding to integer database units since the KLayout Partial tool also rounds to database units.
            new_path = new_path.to_itype(self.layout.dbu).to_dtype(self.layout.dbu)
            self.gui_path = new_path
            self.gui_path_shadow = new_path
        else:
            # Use an empty path to disable editing with the Partial tool
            self.gui_path = pya.DPath()
            self.gui_path_shadow = pya.DPath()

        super().coerce_parameters_impl()

    def build(self):
        """Produce the composite waveguide.

        In practice this becomes an alternating chain of WaveguideCoplanar and some other Element
        subclass. Elements are oriented collinear to the preceding Node and an automatic bend is
        inserted after if the next Node is not collinear.
        """
        self._nodes = Node.nodes_from_string(self.nodes)
        self._child_refpoints = {}
        if len(self._nodes) < 2:
            self.raise_error_on_cell("Need at least 2 Nodes for a WaveguideComposite.",
                                     self._nodes[0].position if len(self._nodes) == 1 else pya.DPoint())

        self._wg_start_idx = 0      # next waveguide starts here
        self._wg_start_pos = self._nodes[0].position
        self._wg_start_dir = self._node_entrance_direction(0)

        for i, node in enumerate(self._nodes):
            self.__log.debug(f' Node #{i}: ({node.position.x:.2f}, {node.position.y:.2f}), {node.element.__class__},'
                             f' {node.params}')
            if node.element is None:
                if 'a' in node.params or 'b' in node.params:
                    self._add_taper(i)
                elif 'face_id' in node.params:
                    self._add_fc_bump(i)
            else:
                if node.element is AirbridgeConnection:
                    self._add_airbridge(i)
                elif issubclass(node.element, Airbridge):
                    if node.element is not Airbridge:   # change default if specific type is used
                        self.airbridge_type = node.element.default_type
                    self._add_airbridge(i, with_side_airbridges=False)
                elif node.element is WaveguideCoplanarTaper:
                    self._add_taper(i)
                elif node.element is FlipChipConnectorRf:
                    self._add_fc_bump(i)
                else:
                    self._add_simple_element(i)

                self._terminator(i)

        # pylint: disable=undefined-loop-variable
        if node.element is None:
            self._add_waveguide(i)

        # Create airbridge on each node that has `ab_across=True` in params
        for i, node in enumerate(self._nodes):
            if "ab_across" in node.params and node.params["ab_across"]:
                ab_len = node.params['bridge_length'] if "bridge_length" in node.params else None
                self._ab_across(node.position - self._node_entrance_direction(i), node.position, 0, ab_len)

        # Reference points
        self.refpoints.update(self._child_refpoints)
        self.add_port("a", self._nodes[0].position, -self._node_entrance_direction(0))
        self.add_port("b", self._wg_start_pos, self._wg_start_dir)


    def _add_taper(self, ind):
        """Create a WaveguideCoplanarTaper and change default a/b."""

        node = self._nodes[ind]

        params = {**self.pcell_params_by_name(WaveguideCoplanarTaper), **node.params}
        a, b = params.pop('a', self.a), params.pop('b', self.b)
        if self.a == a and self.b == b: # no change, just a Node
            return

        taper_cell = self.add_element(WaveguideCoplanarTaper, **{**params, 'a2': a, 'b2': b, 'm2': self.margin})
        self._insert_cell_and_waveguide(ind, taper_cell)

        self.a = a
        self.b = b

    def _add_fc_bump(self, ind):
        """Add FlipChipConnectorRF and change default face_id."""
        node = self._nodes[ind]
        if "face_id" not in node.params:
            node.params["face_id"] = self.face_ids[1]
        params = {**self.pcell_params_by_name(FlipChipConnectorRf), **node.params}
        new_id = node.params.pop("face_id")
        old_id = self.face_ids[0]
        if new_id == old_id:  # no change, just a Node
            return
        # TODO support vias?

        fc_cell = self.add_element(FlipChipConnectorRf, **params)
        self._insert_cell_and_waveguide(ind, fc_cell, before=f'{old_id}_port', after=f'{new_id}_port')

        self.face_ids[0] = new_id
        self.face_ids[1] = old_id

    def _add_airbridge(self, ind, **kwargs):
        """Add an airbridge with tapers at both sides and change default a/b if required."""

        node = self._nodes[ind]
        params = {**self.pcell_params_by_name(AirbridgeConnection), **kwargs,
                  'a2': self.a, 'b2': self.b, 'm2': self.margin,
                  'taper_length': AirbridgeConnection.taper_length,
                  **node.params}

        a, b = params.pop('a', self.a), params.pop('b', self.b)

        if ind == 0:  # set temporary private variables used in _terminator()
            self._ta, self._tb = self.a, self.b

        if {'a', 'b'} & set(node.params):
            params['a2'], params['b2'] = a, b

        if ind == len(self._nodes) - 1:
            self._ta, self._tb = params['a2'], params['b2']

        cell = self.add_element(AirbridgeConnection, **params)
        self._insert_cell_and_waveguide(ind, cell)

        self.a, self.b = a, b

    def _add_simple_element(self, ind):
        """Add any other simple Element, that has port_a and port_b."""

        node = self._nodes[ind]
        params = {**self.pcell_params_by_name(node.element), 'taper_length': self.taper_length, **node.params}

        cell = self.add_element(node.element, **params)
        self._insert_cell_and_waveguide(ind, cell, node.inst_name, *node.align)

    def _insert_cell_and_waveguide(self, ind, cell, inst_name=None, before="port_a", after="port_b"):
        """Place a cell and create the preceding waveguide.

        Normally the element is oriented from the previous node towards this one. Except the first
        one, that goes from here towards the next. The orientation can be also manually fixed by using
        the node parameter `angle`.

        If the element has corner points corresponding to ``before`` and ``after`` (e.g. ``port_a_corner``), these are
        used to determine the entry and exit directions of the ports. If these points do not exist, the waveguides will
        extend the line through ``before`` and ``after``.
        """
        before_corner = before + '_corner'
        after_corner = after + '_corner'

        # Compute cell relative entrance direction
        rel_ref = self.get_refpoints(cell, rec_levels=0)
        if before_corner in rel_ref:
            element_dir = rel_ref[before] - rel_ref[before_corner]
        else:
            element_dir = rel_ref[after] - rel_ref[before]

        # Compute transformation for the cell
        waveguide_dir = self._node_entrance_direction(ind)
        trans = pya.DCplxTrans(1, get_angle(waveguide_dir) - get_angle(element_dir), False, self._nodes[ind].position)
        if ind in (0, len(self._nodes) - 1):
            trans *= pya.DTrans(-rel_ref[before if ind == 0 else after])

        # Insert element cell with computed transformation
        cell_inst, ref = self.insert_cell(cell, trans)
        cell_inst.set_property("waveguide_composite_node_index", ind)
        if inst_name is not None:
            for name, value in ref.items():
                self._child_refpoints[f'{inst_name}_{name}'] = value

        # Add waveguide from previous element until this element
        self._add_waveguide(ind, ref[before], waveguide_dir)

        # Keep track of current position, direction, and index for the future elements
        self._wg_start_pos = ref[after]
        if after_corner in ref:
            _, self._wg_start_dir = vector_length_and_direction(ref[after_corner] - ref[after])
        self._wg_start_idx = ind

    def _node_entrance_direction(self, ind):
        """Returns element entrance direction at node index `ind`."""
        fixed_angle = self._nodes[ind].angle
        if fixed_angle is None:
            prev = max(0, ind - 1)
            return vector_length_and_direction(self._nodes[prev + 1].position - self._nodes[prev].position)[1]
        return get_direction(fixed_angle)

    def _insert_wg_cell(self, points, start_index, end_index):
        """Create and insert waveguide cell.
        Avoid termination if in the middle or if the ends are actual elements.
        """
        if len(points) < 2:
            return
        params = {**self.pcell_params_by_name(WaveguideCoplanar), "path": points}
        if start_index != 0 or self._nodes[start_index].element:
            params['term1'] = 0
        if end_index != len(self._nodes) - 1 or self._nodes[end_index].element:
            params['term2'] = 0
        wg_cell = self.add_element(WaveguideCoplanar, **params)
        cell_inst, _ = self.insert_cell(wg_cell)
        cell_inst.set_property("waveguide_composite_node_index", end_index)

    def _add_waveguide(self, end_index, end_pos=None, end_dir=None):
        """Creates waveguide from `self._wg_start_idx` until `end_index`.

        Args:
            end_index: the last node index taken into account in the waveguide
            end_pos: endpoint position of the waveguide (optional, overwrites `self._nodes[end_index].position`)
            end_dir: endpoint direction of the waveguide (optional)
        """

        def get_corner_lengths(segment_vector, dir_start=pya.DVector(), dir_end=pya.DVector()):
            """Returns distances from segment end points to corner points depending on value of self.tight_routing.
            Returns zero length, if the corner point is not necessary.

            Args:
                segment_vector (pya.DVector): vector from start point to end point
                dir_start (pya.DVector): segment start direction as unit vector (or use zero vector if free direction)
                dir_end (pya.DVector): segment end direction as unit vector (or use zero vector if free direction)

            Returns:
                tuple of lengths
            """
            if not self.tight_routing:
                # Use corner points self.r away from end points.
                _, d = vector_length_and_direction(segment_vector)
                if self.r * abs(d.vprod(dir_start) / 2) < 0.001 and self.r * abs(d.vprod(dir_end) / 2) < 0.001:
                    return 0.0, 0.0
                _, d = vector_length_and_direction(segment_vector - self.r * dir_end)
                if self.r * abs(d.vprod(dir_start) / 2) < 0.001:
                    return 0.0, self.r
                _, d = vector_length_and_direction(segment_vector - self.r * dir_start)
                if self.r * abs(d.vprod(dir_end) / 2) < 0.001:
                    return self.r, 0.0
                return self.r, self.r

            # Use optimal corner routing
            s = segment_vector
            for _ in range(100):  # iterate at most 100 times
                _, d = vector_length_and_direction(s)
                start_len = self.r * abs(d.vprod(dir_start) / (1.0 + d.sprod(dir_start)))
                end_len = self.r * abs(d.vprod(dir_end) / (1.0 + d.sprod(dir_end)))

                # check if converged
                prev_s = s
                s = segment_vector - start_len * dir_start - end_len * dir_end
                if (s - prev_s).length() < 1e-5:
                    return 0.0 if start_len < 0.001 else start_len, 0.0 if end_len < 0.001 else end_len

            # Not converged to up here
            self.raise_error_on_cell("Cannot find suitable routing using 'tight' corners.", self._wg_start_pos)
            return 0.0, 0.0

        # Check if segment has any points
        start_index = self._wg_start_idx
        if end_index <= start_index:
            return

        def curve_length_and_end_points(pnts, p):
            """Returns curve length, curve start point, and curve end point, for given point pnts[p] in list pnts."""
            if p == 0 or p + 1 >= len(pnts):
                return 0.0, pnts[p], pnts[p]
            v1, v2, alpha1, alpha2, _ = WaveguideCoplanar.get_corner_data(pnts[p-1], pnts[p], pnts[p+1], self.r)
            abs_turn = pi - abs(pi - abs(alpha2 - alpha1))
            cut_dist = self.r * tan(abs_turn / 2)
            return self.r * abs_turn, pnts[p] + (-cut_dist / v1.length()) * v1, pnts[p] + (cut_dist / v2.length()) * v2

        # Create waveguide path and create airbridges determined by parameter `n_bridges`.
        points = [self._wg_start_pos]
        straights = {}
        for i in range(start_index, end_index):
            node0 = self._nodes[i]
            node1 = self._nodes[i + 1]

            # Determine segment endpoint positions
            pos0 = self._wg_start_pos if i == start_index else node0.position
            dir0 = self._wg_start_dir if i == start_index else \
                pya.DVector() if node0.angle is None else get_direction(node0.angle)
            pos1 = end_pos if i + 1 == end_index and end_pos is not None else node1.position
            dir1 = end_dir if i + 1 == end_index and end_dir is not None else \
                pya.DVector() if node1.angle is None else get_direction(node1.angle)

            # Add corner points
            len0, len1 = get_corner_lengths(pos1 - pos0, dir0, dir1)
            if len0 > 0:
                points.append(pos0 + len0 * dir0)
            straights[i + 1] = len(points)
            points.append(pos1 + (-len1) * dir1)

            # Add final point if it's not already added
            if i + 1 == end_index and len1 > 0:
                points.append(pos1)

        # Create and insert waveguide cell from points
        # Possibly insert meanders or airbridges on straights
        # Inserting meanders requires splitting of waveguide
        p0 = 0
        n0 = start_index
        point0 = []
        for n1, p1 in straights.items():
            node1 = self._nodes[n1]
            if node1.length_before is not None or node1.length_increment is not None:
                start_len, turn_start, meander_start = curve_length_and_end_points(points, p1-1)
                end_len, meander_end, turn_end = curve_length_and_end_points(points, p1)
                if node1.length_increment is not None:
                    meander_len = (meander_end - meander_start).length() + node1.length_increment
                else:
                    meander_len = node1.length_before
                    if n1 == end_index:
                        meander_len -= end_len + (points[-1] - turn_end).length()
                    elif node1.angle is None:
                        meander_len -= end_len / 2
                    else:
                        meander_len -= end_len + (node1.position - turn_end).length()
                    if n1 - 1 == start_index:
                        meander_len -= start_len + (points[0] - turn_start).length()
                    elif self._nodes[n1-1].angle is None:
                        meander_len -= start_len / 2
                    else:
                        meander_len -= start_len + (self._nodes[n1-1].position - turn_start).length()

                cell_inst, _ = self.insert_cell(Meander, start=[meander_start.x, meander_start.y],
                                                end=[meander_end.x, meander_end.y], length=meander_len, **node1.params)
                cell_inst.set_property("waveguide_composite_node_index", end_index)
                wg_points = point0 + points[p0:p1] + ([] if start_len < 1e-4 else [meander_start])
                self._insert_wg_cell(wg_points, n0, n1)
                n0 = n1
                p0 = p1
                point0 = [] if end_len < 1e-4 else [meander_end]
            elif "n_bridges" in node1.params and node1.params["n_bridges"] > 0:
                ab_len = node1.params['bridge_length'] if "bridge_length" in node1.params else None
                self._ab_across(points[p1-1], points[p1], node1.params["n_bridges"], ab_len)
        self._insert_wg_cell(point0 + points[p0:], n0, end_index)

        # Keep track of current position, direction, and index for the future elements
        self._wg_start_pos = points[-1]
        if self._nodes[end_index].angle is None:
            _, self._wg_start_dir = vector_length_and_direction(points[-1] - points[-2])
        else:
            self._wg_start_dir = get_direction(self._nodes[end_index].angle)
        self._wg_start_idx = end_index

    def _ab_across(self, start, end, num, ab_len=None):
        """Creates ``num`` airbridge crossings equally distributed between ``start`` and ``end``.
        If ``num == 0`` then one airbridge crossing is created at the node. ``ab_len`` is the airbridges' lenght.
        """

        params = {'airbridge_type': self.airbridge_type}
        if ab_len:
            params['bridge_length'] = ab_len
        ab_cell = self.add_element(Airbridge, **params)
        v_dir = end - start
        alpha = get_angle(v_dir)

        if num > 0:
            for i in range(1, num + 1):
                ab_trans = pya.DCplxTrans(1, alpha, False, start + i * v_dir / (num + 1))
                self.insert_cell(ab_cell, ab_trans)
        else:
            ab_trans = pya.DCplxTrans(1, alpha, False, end)
            self.insert_cell(ab_cell, ab_trans)

    def _terminator(self, ind):
        """Terminate the waveguide ending with an Element."""

        if ind == 0 and self.term1 > 0:
            p1 = self._nodes[ind + 1].position
            term = self.term1
        elif ind == len(self._nodes) - 1 and self.term2 > 0:
            p1 = self._nodes[ind - 1].position
            term = self.term2
        else:
            return  # Hasta la vista, baby!

        p2 = self._nodes[ind].position
        a, b = self.a, self.b
        if hasattr(self, '_ta'):
            self.a, self.b = self._ta, self._tb
        WaveguideCoplanar.produce_end_termination(self, p1, p2, term)
        self.a, self.b = a, b


# TODO technical debt: refactor this to be more straightforward and efficient

def produce_fixed_length_bend(element, target_len, point_a, point_a_corner, point_b, point_b_corner, bridges):
    """Inserts a waveguide bend with the given length to the chip.

    Args:
        element: The element to which the waveguide is inserted
        target_len: Target length of the waveguide
        point_a: Endpoint 1 of the waveguide
        point_a_corner: Point towards which the waveguide goes from endpoint 1
        point_b: Endpoint 2 of the waveguide
        point_b_corner: Point towards which the waveguide goes from endpoint 2
        bridges: String determining where airbridges are created, "no" | "middle" | "middle and ends"

    Returns:
        Instance of the created waveguide

    Raises:
        ValueError, if a bend with the given target length and points cannot be created.

    """
    def objective(x):
        return _length_of_var_length_bend(element.layout, element.LIBRARY_NAME, x, point_a, point_a_corner, point_b,
                                          point_b_corner, bridges, element.r) - target_len
    try:
        # floods the database with PCell variants :(
        root = root_scalar(objective, bracket=(element.r, target_len / 2))
        cell = _var_length_bend(element.layout, element.LIBRARY_NAME, root.root, point_a, point_a_corner, point_b,
                                point_b_corner, bridges)
        inst, _ = element.insert_cell(cell)
    except ValueError as e:
        raise ValueError("Cannot create a waveguide bend with length {} between points {} and {}".format(
            target_len, point_a, point_b)) from e

    return inst


def _length_of_var_length_bend(layout, library, corner_dist, point_a, point_a_corner, point_b, point_b_corner,
                               bridges, r):
    # This function shouldn't raise exception, so we have to manually test if waveguide doesn't fit.
    # These tests do not cover all cases, but are enough in most cases
    point_a_shift = point_shift_along_vector(point_a, point_a_corner, corner_dist)
    point_b_shift = point_shift_along_vector(point_b, point_b_corner, corner_dist)
    v1, v2, alpha1, alpha2, _ = WaveguideCoplanar.get_corner_data(point_a, point_a_shift, point_b_shift, r)
    _, v3, _, alpha3, _ = WaveguideCoplanar.get_corner_data(point_a_shift, point_b_shift, point_b, r)
    cut_dist1 = r * tan((pi - abs(pi - abs(alpha2 - alpha1))) / 2)
    cut_dist2 = r * tan((pi - abs(pi - abs(alpha3 - alpha2))) / 2)
    if v1.length() < cut_dist1 or v3.length() < cut_dist2:
        return -1e30  # straight doesn't fit at the ends -> corner_dist is probably too short
    if v2.length() < cut_dist1 + cut_dist2:
        return 1e30  # straight doesn't fit between corners -> corner_dist is probably too large
    b_crosses_a = (v1.vprod_sign(point_b - point_a) * v1.vprod_sign(point_b_shift - point_a)) < 0
    a_crosses_b = (v3.vprod_sign(point_a - point_b) * v3.vprod_sign(point_a_shift - point_b)) < 0
    if b_crosses_a and a_crosses_b:
        return 1e30  # waveguide is crossing itself -> corner_dist is probably too large

    # Create waveguide and measure it's length
    cell = _var_length_bend(layout, library, corner_dist, point_a, point_a_corner, point_b, point_b_corner, bridges)
    return get_cell_path_length(cell)


def _var_length_bend(layout, library, corner_dist, point_a, point_a_corner, point_b, point_b_corner, bridges):
    cell = WaveguideComposite.create(layout, library, nodes=[
        Node(point_a, ab_across=bridges.endswith("ends")),
        Node(point_shift_along_vector(point_a, point_a_corner, corner_dist)),
        Node(point_shift_along_vector(point_b, point_b_corner, corner_dist), n_bridges=bridges.startswith("middle")),
        Node(point_b, ab_across=bridges.endswith("ends")),
    ])
    return cell



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from math import sqrt
from autologging import logged
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.junctions.squid import Squid
from kqcircuits.util.symmetric_polygons import polygon_with_vsym


@logged
class Manhattan(Squid):
    """The PCell declaration for a Manhattan style SQUID.

    This SQUID has two distinct sub-types automatically selected by loop-area.
    """

    finger_overshoot = Param(pdt.TypeDouble, "Length of fingers after the junction.", 1.0, unit="μm")
    include_base_metal_gap = Param(pdt.TypeBoolean, "Include base metal gap layer", True)
    shadow_margin = Param(pdt.TypeDouble, "Shadow layer margin near the the pads", 1.0, unit="μm")
    compact_geometry = Param(pdt.TypeBoolean, "Compact geometry for metal addition.", False)
    separate_junctions = Param(pdt.TypeBoolean, "Junctions to separate layer.", False)
    offset_compensation = Param(pdt.TypeDouble, "Junction lead offset from junction width", 0, unit="μm")
    mirror_offset = Param(pdt.TypeBoolean, "Move the junction lead offset to the other lead", False)
    finger_overlap = Param(pdt.TypeDouble, "Length of fingers inside the pads.", 0.2, unit="μm")

    def build(self):
        self.produce_manhattan_squid(top_pad_layer="SIS_junction")

    def produce_manhattan_squid(self, top_pad_layer):

        # geometry constants
        big_loop_height = 10
        loop_bottom_y = 1.5
        self.metal_gap_top_y = 20 if self.compact_geometry else 26.5
        self.width = 36 if self.compact_geometry else 38  # total width of junction layer
        self.height = 17 if self.compact_geometry else 20.2  # total height of junction layer
        bp_height = 5  # bottom pad height
        tp_width = 10  # top pad width
        brim_height = 1  # thickness of the "top-hat's" brim
        small_loop_height = 5.2
        small_hat_width = 2  # width of the small hat shape, in case of small loop

        # corner rounding parameters
        rounding_params = {
            "rinner": 0.5,  # inner corner rounding radius
            "router": 0.5,  # outer corner rounding radius
            "n": 64,  # number of point per rounded corner
        }

        # convenience variables
        delta_j = self.loop_area / big_loop_height  # junction distance, a.k.a. loop width
        tp_height = self.height - loop_bottom_y - big_loop_height  # top pad height
        bp_gap_x = -self.width / 2 + (self.width - delta_j) / 2  # bottom gap left edge x-coordinate
        bp_gap_x_min = -self.width / 2 + 7  # fixed at minimum size
        finger_margin = brim_height  # make hats brim this much wider for good finger connection

        # adjust for small loop geometry
        small_loop = tp_width > -bp_gap_x * 2
        if small_loop:
            bp_gap_x = bp_gap_x_min
            delta_j = self.loop_area / small_loop_height

        junction_shapes_top = []
        junction_shapes_bottom = []
        shadow_shapes = []

        # create rounded bottom part and top parts
        self.produce_contact_pads(top_pad_layer, bp_height, bp_gap_x, tp_height, tp_width,
                                  big_loop_height, junction_shapes_bottom, rounding_params,
                                  shadow_shapes, junction_shapes_top)

        # create rectangular junction-support structures and junctions
        if small_loop:
            small_hat = [
                pya.DPoint(-small_hat_width / 2, self.height - tp_height),
                pya.DPoint(-small_hat_width / 2, small_loop_height + loop_bottom_y + brim_height),
                pya.DPoint(-delta_j / 2 - finger_margin, small_loop_height + loop_bottom_y + brim_height),
                pya.DPoint(-delta_j / 2 - finger_margin, small_loop_height + loop_bottom_y)
            ]
            junction_shapes_top.append(polygon_with_vsym(small_hat).to_itype(self.layout.dbu))
            if top_pad_layer != "SIS_junction":
                junction_shapes_bottom.append(polygon_with_vsym(small_hat).to_itype(self.layout.dbu))
            small_hat_shadow = [
                small_hat[0] + pya.DPoint(-self.shadow_margin, -self.shadow_margin),
                small_hat[1] + pya.DPoint(-self.shadow_margin, self.shadow_margin),
                small_hat[2] + pya.DPoint(-self.shadow_margin, self.shadow_margin),
                small_hat[3] + pya.DPoint(-self.shadow_margin, -self.shadow_margin),
            ]
            shadow_shapes.append(polygon_with_vsym(small_hat_shadow).to_itype(self.layout.dbu))
            small_hat[3].x += finger_margin
            self._make_junctions(small_hat[3], loop_bottom_y)
        else:
            tp_brim_left = [
                pya.DPoint(-delta_j / 2 - finger_margin, self.height - tp_height + brim_height),
                pya.DPoint(-delta_j / 2 - finger_margin, self.height - tp_height)
            ]
            junction_shapes_top.append(polygon_with_vsym(tp_brim_left).to_itype(self.layout.dbu))
            if top_pad_layer != "SIS_junction":
                junction_shapes_bottom.append(polygon_with_vsym(tp_brim_left).to_itype(self.layout.dbu))
            tp_brim_shadow_pts = [
                tp_brim_left[0] + pya.DPoint(-self.shadow_margin, self.shadow_margin),
                tp_brim_left[1] + pya.DPoint(-self.shadow_margin, -self.shadow_margin),
            ]
            shadow_shapes.append(polygon_with_vsym(tp_brim_shadow_pts).to_itype(self.layout.dbu))
            tp_brim_left[1].x += finger_margin
            self._make_junctions(tp_brim_left[1], bp_height, finger_margin)

        self._add_shapes(junction_shapes_bottom, "SIS_junction")
        self._add_shapes(junction_shapes_top, top_pad_layer)
        self._add_shapes(shadow_shapes, "SIS_shadow")
        self._produce_ground_grid_avoidance()
        self._produce_ground_metal_shapes()
        self._add_refpoints()

    def produce_contact_pads(self,top_pad_layer, bp_height, bp_gap_x, tp_height,tp_width, big_loop_height,
                             junction_shapes_bottom, rounding_params, shadow_shapes, junction_shapes_top):

        bp_pts_left = [
            pya.DPoint(-self.width / 2, -0.5),
            pya.DPoint(-self.width / 2, bp_height),
            pya.DPoint(bp_gap_x, bp_height),
            pya.DPoint(bp_gap_x, self.height - tp_height - big_loop_height)
        ]
        bp_shape = polygon_with_vsym(bp_pts_left)
        self._round_corners_and_append(bp_shape, junction_shapes_bottom, rounding_params)

        bp_shadow_pts_left = [
            bp_pts_left[0] + pya.DPoint(-self.shadow_margin, -self.shadow_margin),
            bp_pts_left[1] + pya.DPoint(-self.shadow_margin, self.shadow_margin),
            bp_pts_left[2] + pya.DPoint(self.shadow_margin, self.shadow_margin),
            bp_pts_left[3] + pya.DPoint(self.shadow_margin, self.shadow_margin),
        ]
        bp_shadow_shape = polygon_with_vsym(bp_shadow_pts_left)
        self._round_corners_and_append(bp_shadow_shape, shadow_shapes, rounding_params)

        # create rounded top part

        tp_pts_left = [
            pya.DPoint(-tp_width / 2, self.height),
            pya.DPoint(-tp_width / 2, self.height - tp_height),
        ]
        tp_shape = polygon_with_vsym(tp_pts_left)
        self._round_corners_and_append(tp_shape, junction_shapes_top, rounding_params)

        # add top pad to bottom shapes in case another layer is used for the upper part of the squid

        if top_pad_layer != "SIS_junction":
            self._round_corners_and_append(tp_shape, junction_shapes_bottom, rounding_params)

        tp_shadow_pts_left = [
            tp_pts_left[0] + pya.DPoint(-self.shadow_margin, self.shadow_margin),
            tp_pts_left[1] + pya.DPoint(-self.shadow_margin, -self.shadow_margin),
        ]
        tp_shadow_shape = polygon_with_vsym(tp_shadow_pts_left)
        self._round_corners_and_append(tp_shadow_shape, shadow_shapes, rounding_params)

    def _make_junctions(self, top_corner, b_corner_y, finger_margin=0):
        """Create junction fingers and add them to some SIS layer.

        Choose 'SIS_junction' layer by default but 'SIS_junction_2' if ``separate_junctions`` is True.
        """
        jx = top_corner.x - (top_corner.y - b_corner_y) / 2
        jy = (top_corner.y + b_corner_y) / 2
        ddb = self.junction_width * sqrt(0.5)
        ddt = self.junction_width * sqrt(0.5)
        if self.mirror_offset:
            ddt += self.offset_compensation * sqrt(0.5)
        else:
            ddb += self.offset_compensation * sqrt(0.5)
        fo = self.finger_overshoot * sqrt(0.5)
        pl = self.finger_overlap * sqrt(0.5)  # plus length to connect despite of rounding

        def finger_points(size):
            return [
                pya.DPoint(top_corner.x + pl, top_corner.y + size + pl),
                pya.DPoint(top_corner.x + size + pl, top_corner.y + pl),
                pya.DPoint(jx - fo, jy - fo - size),
                pya.DPoint(jx - fo - size, jy - fo),
            ]

        finger_bottom = pya.DTrans(-jx, -jy) * pya.DPolygon(finger_points(ddb))
        finger_top = pya.DTrans(-jx, -jy) * pya.DPolygon(finger_points(ddt))

        junction_shapes = [(pya.DTrans(jx - finger_margin, jy) * finger_top).to_itype(self.layout.dbu),
                           (pya.DTrans(0, False, jx - 2 * top_corner.x, jy) * finger_top).to_itype(self.layout.dbu),
                           (pya.DTrans(3, False, jx - finger_margin, jy) * finger_bottom).to_itype(self.layout.dbu),
                           (pya.DTrans(3, False, jx - 2 * top_corner.x, jy) * finger_bottom).to_itype(self.layout.dbu)]

        junction_region = pya.Region(junction_shapes).merged()
        layer_name = "SIS_junction_2" if self.separate_junctions else "SIS_junction"
        self.cell.shapes(self.get_layer(layer_name)).insert(junction_region)

        # place refpoints at the middle of the left and right junctions
        squa = sqrt(2) / 2
        self.refpoints["l"] = pya.DPoint(jx - fo - finger_margin + self.finger_overshoot * squa,
                                         jy - fo + self.finger_overshoot * squa)
        self.refpoints["r"] = pya.DPoint(jx - fo - 2 * top_corner.x + self.finger_overshoot * squa,
                                         jy - fo + self.finger_overshoot * squa)

    def _add_shapes(self, shapes, layer):
        """Merge shapes into a region and add it to layer."""
        region = pya.Region(shapes).merged()
        self.cell.shapes(self.get_layer(layer)).insert(region)

    def _add_refpoints(self):
        """Adds the "origin_squid" refpoint and port "common"."""
        self.refpoints["origin_squid"] = pya.DPoint(0, 0)
        self.add_port("common", pya.DPoint(0, self.metal_gap_top_y))

    def _produce_ground_metal_shapes(self):
        """Produces hardcoded shapes in metal gap and metal addition layers."""
        # metal additions bottom
        x0 = -12 if self.compact_geometry else -13
        y0 = -1
        bottom_pts = [
            pya.DPoint(x0 - 3, y0 - 1),
            pya.DPoint(x0 - 3, y0 + 2),
            pya.DPoint(x0 - 5, y0 + 2),
            pya.DPoint(x0 - 5, y0 + 5),
            pya.DPoint(x0, y0 + 5),
            pya.DPoint(x0, y0 + 1)
        ]
        shape = polygon_with_vsym(bottom_pts)
        self.cell.shapes(self.get_layer("base_metal_addition")).insert(shape)
        # metal additions top
        y0 = 12 if self.compact_geometry else 14.5
        top_pts = [
            pya.DPoint(-2, y0 + 3),
            pya.DPoint(-2, y0 + 1),
            pya.DPoint(-1, y0 + 1),
            pya.DPoint(-1, y0),
            pya.DPoint(-4, y0),
            pya.DPoint(-4, self.metal_gap_top_y),
        ]
        shape = polygon_with_vsym(top_pts)
        self.cell.shapes(self.get_layer("base_metal_addition")).insert(shape)
        # metal gap
        if self.include_base_metal_gap:
            pts = bottom_pts[::-1] + [pya.DPoint(-20.5, -2), pya.DPoint(-20.5, self.metal_gap_top_y)] + top_pts[::-1]
            shape = polygon_with_vsym(pts)
            self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)

    def _produce_ground_grid_avoidance(self):
        """Add ground grid avoidance."""
        w = self.cell.dbbox().width()
        h = self.cell.dbbox().height()
        protection = pya.DBox(-w / 2 - self.margin, -2 - self.margin, w / 2 + self.margin, h - 2 + self.margin)
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(protection)

    def _round_corners_and_append(self, polygon, polygon_list, rounding_params):
        """Rounds the corners of the polygon, converts it to integer coordinates, and adds it to the polygon list."""
        polygon = polygon.round_corners(rounding_params["rinner"], rounding_params["router"], rounding_params["n"])
        polygon_list.append(polygon.to_itype(self.layout.dbu))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import math
import numpy
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.tsvs.tsv import Tsv


class TsvEllipse(Tsv):
    """Connector between faces of two sides of a substrate.

    Origin is at the geometric center. Geometry is elliptical.

   .. MARKERS_FOR_PNG -0.2,0
    """

    tsv_elliptical_width = Param(pdt.TypeDouble, "TSV elliptical width", 30, unit="μm")

    def produce_impl(self):
        self.create_tsv_connector()

    def create_tsv_connector(self):
        """
        Generate elliptical TSV
        """
        # shorthand
        r = self.tsv_diameter / 2
        w = self.tsv_elliptical_width / 2
        m = self.margin

        # parametric representation is taken from https://en.wikipedia.org/wiki/Superellipse
        p1 = 6
        p2 = 2
        # Protection layer
        tsv_pts_avoidance = [pya.DPoint(
            numpy.abs(math.cos(a)) ** (2 / p1) * (w + m) * numpy.sign(math.cos(a)),
            numpy.abs(math.sin(a)) ** (2 / p2) * (r + m) * numpy.sign(math.sin(a))) for a in
            (x / 32 * math.pi for x in range(0, 65))]

        tsv_pts = [
            pya.DPoint(numpy.abs(math.cos(a)) ** (2 / p1) * w * numpy.sign(math.cos(a)),
                       numpy.abs(math.sin(a)) ** (2 / p2) * r * numpy.sign(math.sin(a))) for
            a in (x / 32 * math.pi for x in range(0, 65))]

        shape = pya.DPolygon(tsv_pts_avoidance)
        # ground avoidance layer 1t1 face
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(shape)
        self.cell.shapes(self.get_layer("ground_grid_avoidance", 1)).insert(shape)
        self.cell.shapes(self.get_layer("through_silicon_via")).insert(pya.DPolygon(tsv_pts))
        self.cell.shapes(self.get_layer("through_silicon_via", 1)).insert(pya.DPolygon(tsv_pts))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.chips.chip import Chip
from kqcircuits.defaults import default_airbridge_type
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.airbridge_connection import AirbridgeConnection
from kqcircuits.util.coupler_lib import cap_params
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node


class QualityFactor(Chip):
    """The PCell declaration for a QualityFactor chip."""
    res_lengths = Param(pdt.TypeList, "Resonator lengths", [5434, 5429, 5374, 5412, 5493, 5589], unit="[μm]",
                        docstring="Physical length of resonators [μm]")
    n_fingers = Param(pdt.TypeList, "Number of fingers of the coupler", [4, 4, 2, 4, 4, 4],
                      docstring="Fingers in planar capacitors")
    l_fingers = Param(pdt.TypeList, "Length of fingers", [23.1, 9.9, 14.1, 10, 21, 28], unit="[μm]",
                      docstring="Length of the capacitor fingers [μm]")
    type_coupler = Param(pdt.TypeList, "Coupler types",
                         ["interdigital", "interdigital", "interdigital", "gap", "gap", "gap"])
    n_ab = Param(pdt.TypeList, "Number of resonator airbridges", [5, 0, 5, 5, 5, 5])
    res_term = Param(pdt.TypeList, "Resonator termination type",
        ["galvanic", "galvanic", "galvanic", "airbridge", "airbridge", "airbridge"])
    res_beg = Param(pdt.TypeList, "Resonator beginning type",
        ["galvanic", "galvanic", "galvanic", "airbridge", "airbridge", "airbridge"])
    res_a = Param(pdt.TypeList, "Resonator waveguide center conductor width", [5, 10, 20, 5, 10, 20], unit="[μm]",
                  docstring="Width of the center conductor in the resonators [μm]")
    res_b = Param(pdt.TypeList, "Resonator waveguide gap width", [3, 6, 12, 3, 6, 12], unit="[μm]",
                  docstring="Width of the gap in the resonators [μm]")
    tl_airbridges = Param(pdt.TypeBoolean, "Airbridges on transmission line", True)
    res_airbridge_types = Param(pdt.TypeList, "Airbridge type for each resonator",
                         default=[default_airbridge_type]*6)
    launcher_top_dist = Param(pdt.TypeDouble, "Launcher distance from top", 2800, unit="μm")
    launcher_indent = Param(pdt.TypeDouble, "Launcher indentation from edge", 800, unit="μm")
    marker_safety = Param(pdt.TypeDouble, "Distance between launcher and first curve", 1000, unit="μm")
    feedline_bend_distance = Param(pdt.TypeDouble, "Horizontal distance of feedline bend", 100, unit="μm")
    resonators_both_sides = Param(pdt.TypeBoolean, "Place resonators on both sides of feedline", False)
    max_res_len = Param(pdt.TypeDouble, "Maximal straight length of resonators", 1e30, unit="μm",
                        docstring="Resonators exceeding this length become meandering")
    # override box to have hidden=False and allow GUI editing
    box = Param(pdt.TypeShape, "Border", pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000)))

    def build(self):
        # Interpretation of parameter lists
        res_lengths = [float(foo) for foo in self.res_lengths]
        res_a = [float(foo) for foo in self.res_a]
        res_b = [float(foo) for foo in self.res_b]
        n_fingers = [float(foo) for foo in self.n_fingers]
        type_coupler = self.type_coupler
        n_ab = [int(foo) for foo in self.n_ab]
        l_fingers = [float(foo) for foo in self.l_fingers]
        res_term = self.res_term
        res_beg = self.res_beg

        # center the resonators in the chip regardless of size
        max_res_len = min(max(res_lengths), self.max_res_len)
        chip_side = self.box.p2.y - self.box.p1.y
        if self.resonators_both_sides:
            wg_top_y = chip_side / 2
        else:
            wg_top_y = (chip_side + max_res_len) / 2

        # Non-standard Launchers mimicking SMA8 at 1cm chip size, but keeping fixed distance from top
        launchers = self.produce_n_launchers(8, "RF", 300, 180, self.launcher_indent,
                                             chip_side - 2 * self.launcher_top_dist, {8: "PL-IN", 3: "PL-OUT"})

        # Define start and end of feedline
        points_fl = [launchers["PL-IN"][0]]
        if abs(launchers["PL-IN"][0].y - wg_top_y) > 1:
            # Bend in the feedline needed
            points_fl += [
                launchers["PL-IN"][0] + pya.DVector(self.r + self.marker_safety, 0),
                pya.DPoint(launchers["PL-IN"][0].x + self.r + self.feedline_bend_distance + self.marker_safety,
                           wg_top_y)
            ]
            points_fl_end = [
                pya.DPoint(launchers["PL-OUT"][0].x - self.r - self.feedline_bend_distance - self.marker_safety,
                           wg_top_y),
                launchers["PL-OUT"][0] + pya.DVector(-self.r - self.marker_safety, 0),
            ]
        elif self.marker_safety > 0:
            points_fl += [launchers["PL-IN"][0] + pya.DVector(self.marker_safety, 0)]
            points_fl_end = [
                launchers["PL-OUT"][0] + pya.DVector(-self.marker_safety, 0),
            ]
        else:
            points_fl_end = []

        points_fl_end += [launchers["PL-OUT"][0]]

        tl_start = points_fl[-1]
        tl_end = points_fl_end[0]

        resonators = len(self.res_lengths)
        v_res_step = (tl_end - tl_start) * (1. / resonators)
        cell_cross = self.add_element(WaveguideCoplanarSplitter, **t_cross_parameters(
            length_extra_side=2 * self.a, a=self.a, b=self.b, a2=self.a, b2=self.b))

        # Airbridge crossing resonators
        cell_ab_crossing = self.add_element(Airbridge)

        for i in range(resonators):
            resonator_up = self.resonators_both_sides and (i % 2) == 0

            # Cross
            cross_trans = pya.DTrans(0, resonator_up, tl_start + v_res_step * (i + 0.5))
            _, cross_refpoints_abs = self.insert_cell(cell_cross, cross_trans)

            # Coupler
            _, cplr_refpoints_abs = self.insert_cell(
                trans=pya.DTrans.R270 if resonator_up else pya.DTrans.R90,
                align="port_b",
                align_to=cross_refpoints_abs["port_bottom"],
                **cap_params(n_fingers[i], l_fingers[i], type_coupler[i], element_key='cell',
                             a=res_a[i], b=res_b[i], a2=self.a, b2=self.b)
            )

            pos_res_start = cplr_refpoints_abs["port_a"]
            sign = 1 if resonator_up else -1
            pos_res_end = pos_res_start + sign*pya.DVector(0, min(res_lengths[i], self.max_res_len))
            self.refpoints['resonator_{}_end'.format(i)] = pos_res_end

            # create resonator using WaveguideComposite
            if res_beg[i] == "airbridge":
                node_beg = Node(pos_res_start, AirbridgeConnection, with_side_airbridges=False)
            else:
                node_beg = Node(pos_res_start)

            length_increment = res_lengths[i] - self.max_res_len if res_lengths[i] > self.max_res_len else None
            bridge_length = res_a[i] + 2 * res_b[i] + 38
            if res_term[i] == "airbridge":
                node_end = Node(pos_res_end, AirbridgeConnection, with_side_airbridges=False,
                                with_right_waveguide=False, n_bridges=n_ab[i],
                                bridge_length=bridge_length, length_increment=length_increment)
            else:
                node_end = Node(pos_res_end, n_bridges=n_ab[i],
                                bridge_length=bridge_length, length_increment=length_increment)

            airbridge_type = default_airbridge_type
            if i < len(self.res_airbridge_types):
                airbridge_type = self.res_airbridge_types[i]
            wg = self.add_element(WaveguideComposite, nodes=[node_beg, node_end], a=res_a[i], b=res_b[i],
                airbridge_type=airbridge_type)
            self.insert_cell(wg)

            # Feedline
            self.insert_cell(WaveguideCoplanar, **{**self.cell.pcell_parameters_by_name(), **{
                "path": pya.DPath(points_fl + [
                    cross_refpoints_abs["port_left"]
                ], 1),
                "term2": 0
            }})
            points_fl = [cross_refpoints_abs["port_right"]]

            # airbridges on the left and right side of the couplers
            if self.tl_airbridges:
                ab_dist_to_coupler = 60.0
                ab_coupler_left = pya.DPoint((cross_refpoints_abs["port_left"].x) - ab_dist_to_coupler,
                                             (cross_refpoints_abs["port_left"].y))
                ab_coupler_right = pya.DPoint((cross_refpoints_abs["port_right"].x) + ab_dist_to_coupler,
                                              (cross_refpoints_abs["port_right"].y))

                self.insert_cell(cell_ab_crossing, pya.DTrans(0, False, ab_coupler_left))
                self.insert_cell(cell_ab_crossing, pya.DTrans(0, False, ab_coupler_right))

        # Last feedline

        self.insert_cell(WaveguideCoplanar, **{**self.cell.pcell_parameters_by_name(), **{
            "path": pya.DPath(points_fl + points_fl_end, 1),
            "term2": 0
        }})



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import logging
from string import Template


def apply_template(filename_template, filename_output, rules):
    with open(filename_template) as filein:
        src = Template(filein.read())
    results = src.substitute(rules)
    with open(filename_output, "w") as fileout:
        fileout.write(results)
    # dirname_sondata = os.path.join(os.path.dirname(filename_output), "sondata")
    # if not os.path.exists(dirname_sondata):
    #     os.mkdir(dirname_sondata)
    # dirname_project = os.path.join(dirname_sondata, os.path.splitext(os.path.basename(filename_output))[0])
    # if not os.path.exists(dirname_project):
    #     os.mkdir(dirname_project)


def polygon_head(
        nvertices,  # number of vertices of the polygon
        debugid,  # unique number for sonnet internal debugging
        ilevel=0,  # sonnet layer number
        mtype=-1,  # metallization type index, -1 for lossless
        filltype="N",  # N for staircase, T for diagonal, V for conformal
        xmin=1,  # minimum subsection size
        ymin=1,  # minimum subsection size
        xmax=100,  # maximum subsection size
        ymax=100,  # maximum subsection size
        conmax=0,  # maximum length for conformal mesh subsection, 0 for auto
        res=0,  # reserved for sonnet future
        edge_mesh="Y"  # edge mesh on (Y) or off (N)
):
    return f"{ilevel} {nvertices} {mtype} {filltype} {debugid} {xmin} {ymin} {xmax} {ymax} {conmax} {res} {res} " \
           f"{edge_mesh}\n"


def symmetry(sym: bool = False):
    sonnet_str = ""
    if sym:
        sonnet_str = "SYM"
    return sonnet_str


def box(
        xwidth: float = 8000.,
        ywidth: float = 8000.,
        xcells: int = 8000,
        ycells: int = 8000,
        materials_type: str = "Si BT"
):
    xcells2 = 2 * xcells
    ycells2 = 2 * ycells
    nsubs = 20  # placeholder for deprecated parameter
    eeff = 0  # placeholder for deprecated parameter

    materials = {
        "Si RT": "3000 1 1 0 0 0 0 \"vacuum\"\n500 11.7 1 0 0 0 0 \"Silicon (room temperature)\"",
        "Si BT": "3000 1 1 0 0 0 0 \"vacuum\"\n500 11.45 1 1e-006 0 0 0 \"Silicon (10mK)\"",
        "SiOx+Si": "3000 1 1 0 0 0 0 \"vacuum\"\n0.55 3.78 11.7 1 0 0 0 \"SiOx (10mK)\"\n525 11.45 1 1e-06 0 0 0 \"Si "
                   "(10mK)\"",
        "Si+Al": "3000 1 1 0 0 0 0 \"vacuum\"\n0.5 9.9 1 0.0001 0 0 0 \"Alumina (99.5%)\"\n0.45 1 1 0 0 0 0 \"vacuum\""
                 "\n525 11.45 1 1e-06 0 0 0 \"Si (10mK)\"",
    }[materials_type]

    nlev = {
        "Si": 1,
        "Si BT": 1,
        "SiOx+Si": 2,
        "Si+Al": 3
    }[materials_type]

    return f"BOX {nlev} {xwidth} {ywidth} {xcells2} {ycells2} {nsubs} {eeff}\n{materials}"


def refplane(
        position: str,  # "LEFT" | "RIGHT" | "TOP" | "BOTTOM",
        length: int = 0,
        port_ipoly: str = "" # "LINK" or "FIX"
        ):
    if port_ipoly != "":
        plane_type = "LINK"
        poly = "POLY {} 1\n0\n".format(port_ipoly[0])
        length = ""
    else:
        plane_type = "FIX"
        poly = ""
    return f"DRP1 {position} {plane_type} {length}\n{poly}"


def refplanes(positions, length, port_ipolys):
    sonnet_str = ""
    for i, pos in enumerate(positions):
        sonnet_str += refplane(pos, length, port_ipolys[i])
    return sonnet_str


def port(
        portnum,
        ipolygon,
        ivertex,
        port_type="STD",  # STD for standard | AGND autogrounded | CUP cocalibrated
        xcord=0,  # pylint: disable=unused-argument
        ycord=0,  # pylint: disable=unused-argument
        group="",
        resist=50,
        react=0,
        induct=0,
        capac=0
):
    if group:
        group = '"' + group + '"'
    logging.info(locals())
    return f"POR1 {port_type} {group}\nPOLY {ipolygon} 1\n{ivertex}\n{portnum} {resist} {react} {induct} {capac}\n"
    # {xcord} {ycord} [reftype rpcallen]


# def ports(shapes):
#  sonnet_str = ""
#  polygons = 0
#
#  # FIXME Maybe the shapes will not have the same indexes as polygons in the region!
#  for shape in shapes.each():
#    if shape:
#      polygons += 1
#      ivertex = shape.property("sonnet_port_edge")
#      portnum = shape.property("sonnet_port_nr")
#      if ivertex!=None and portnum!=None:
#        sonnet_str += port(ipolygon=polygons-1, portnum=portnum, ivertex=ivertex)
#
#  return sonnet_str

def control(control_type):
    return {
        "Simple": "SIMPLE",  # Linear frequency sweep
        "ABS": "ABS", # Sonnet guesses the resonances, simulates about 5 points around the resonance and interpolates
                      # the rest
        "Sweep": "VARSWP"
    }[control_type]


def polygons(polygons, v, dbu, ilevel, fill_type):
    sonnet_str = 'NUM {}\n'.format(len(polygons))
    for i, hole_poly in enumerate(polygons):
        poly = hole_poly.resolved_holes()

        if hasattr(poly, 'isVia'):
            sonnet_str += via(poly, debugid=i, ilevel=next(ilevel))
        else:
            sonnet_str += polygon_head(nvertices=poly.num_points_hull() + 1,
                                   debugid=i + 1, ilevel=next(ilevel),
                                   filltype=fill_type)  # "Debugid" is actually used for mapping ports to polygons, 0 is
                                                        # not allowed

        for _, point in enumerate(poly.each_point_hull()):
            sonnet_str += "{} {}\n".format(point.x * dbu + v.x,
                                           -(point.y * dbu + v.y))  # sonnet Y-coordinate goes in the other direction
        point = next(poly.each_point_hull())  # first point again to close the polygon
        sonnet_str += "{} {}\nEND\n".format(point.x * dbu + v.x, -(point.y * dbu + v.y))

    return sonnet_str


def via(poly, debugid, ilevel):
    via_head = polygon_head(nvertices=poly.num_points_hull() + 1, debugid=debugid, ilevel=ilevel, mtype=0)
    return "VIA POLYGON\n" + via_head + "TOLEVEL 1 RING COVERS\n"
